import uuid
from datetime import date, time, datetime, timedelta
from decimal import Decimal

from hazelcast import six
from hazelcast.six.moves import range
from hazelcast.protocol.client_message import (
    NULL_FRAME_BUF,
    BEGIN_FRAME_BUF,
    END_FRAME_BUF,
    SIZE_OF_FRAME_LENGTH_AND_FLAGS,
    _IS_FINAL_FLAG,
    NULL_FINAL_FRAME_BUF,
    END_FINAL_FRAME_BUF,
)
from hazelcast.serialization.bits import (
    LONG_SIZE_IN_BYTES,
    UUID_SIZE_IN_BYTES,
    LE_INT,
    LE_LONG,
    BOOLEAN_SIZE_IN_BYTES,
    INT_SIZE_IN_BYTES,
    LE_ULONG,
    LE_UINT16,
    LE_INT8,
    UUID_MSB_SHIFT,
    UUID_LSB_MASK,
    BYTE_SIZE_IN_BYTES,
    SHORT_SIZE_IN_BYTES,
    LE_INT16,
    FLOAT_SIZE_IN_BYTES,
    LE_FLOAT,
    LE_DOUBLE,
    DOUBLE_SIZE_IN_BYTES,
)
from hazelcast.serialization.data import Data
from hazelcast.util import int_from_bytes, timezone

_LOCAL_DATE_SIZE_IN_BYTES = SHORT_SIZE_IN_BYTES + BYTE_SIZE_IN_BYTES * 2
_LOCAL_TIME_SIZE_IN_BYTES = BYTE_SIZE_IN_BYTES * 3 + INT_SIZE_IN_BYTES
_LOCAL_DATE_TIME_SIZE_IN_BYTES = _LOCAL_DATE_SIZE_IN_BYTES + _LOCAL_TIME_SIZE_IN_BYTES
_OFFSET_DATE_TIME_SIZE_IN_BYTES = _LOCAL_DATE_TIME_SIZE_IN_BYTES + INT_SIZE_IN_BYTES


class CodecUtil(object):
    @staticmethod
    def fast_forward_to_end_frame(msg):
        # We are starting from 1 because of the BEGIN_FRAME we read
        # in the beginning of the decode method
        num_expected_end_frames = 1
        while num_expected_end_frames != 0:
            frame = msg.next_frame()
            if frame.is_end_frame():
                num_expected_end_frames -= 1
            elif frame.is_begin_frame():
                num_expected_end_frames += 1

    @staticmethod
    def encode_nullable(buf, value, encoder, is_final=False):
        if value is None:
            if is_final:
                buf.extend(NULL_FINAL_FRAME_BUF)
            else:
                buf.extend(NULL_FRAME_BUF)
        else:
            encoder(buf, value, is_final)

    @staticmethod
    def decode_nullable(msg, decoder):
        if CodecUtil.next_frame_is_null_frame(msg):
            return None
        else:
            return decoder(msg)

    @staticmethod
    def next_frame_is_data_structure_end_frame(msg):
        return msg.peek_next_frame().is_end_frame()

    @staticmethod
    def next_frame_is_null_frame(msg):
        is_null = msg.peek_next_frame().is_null_frame()
        if is_null:
            msg.next_frame()
        return is_null


class ByteArrayCodec(object):
    @staticmethod
    def encode(buf, value, is_final=False):
        header = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
        LE_INT.pack_into(header, 0, SIZE_OF_FRAME_LENGTH_AND_FLAGS + len(value))
        if is_final:
            LE_UINT16.pack_into(header, INT_SIZE_IN_BYTES, _IS_FINAL_FLAG)
        buf.extend(header)
        buf.extend(value)

    @staticmethod
    def decode(msg):
        return msg.next_frame().buf


class DataCodec(object):
    @staticmethod
    def encode(buf, value, is_final=False):
        value_bytes = value.to_bytes()
        header = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
        LE_INT.pack_into(header, 0, SIZE_OF_FRAME_LENGTH_AND_FLAGS + len(value_bytes))
        if is_final:
            LE_UINT16.pack_into(header, INT_SIZE_IN_BYTES, _IS_FINAL_FLAG)
        buf.extend(header)
        buf.extend(value_bytes)

    @staticmethod
    def decode(msg):
        return Data(msg.next_frame().buf)

    @staticmethod
    def encode_nullable(buf, value, is_final=False):
        if value is None:
            if is_final:
                buf.extend(NULL_FINAL_FRAME_BUF)
            else:
                buf.extend(NULL_FRAME_BUF)
        else:
            DataCodec.encode(buf, value, is_final)

    @staticmethod
    def decode_nullable(msg):
        if CodecUtil.next_frame_is_null_frame(msg):
            return None
        else:
            return DataCodec.decode(msg)


class EntryListCodec(object):
    @staticmethod
    def encode(buf, entries, key_encoder, value_encoder, is_final=False):
        buf.extend(BEGIN_FRAME_BUF)
        for key, value in entries:
            key_encoder(buf, key)
            value_encoder(buf, value)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def encode_nullable(buf, entries, key_encoder, value_encoder, is_final=False):
        if entries is None:
            if is_final:
                buf.extend(NULL_FINAL_FRAME_BUF)
            else:
                buf.extend(NULL_FRAME_BUF)
        else:
            EntryListCodec.encode(buf, entries, key_encoder, value_encoder, is_final)

    @staticmethod
    def decode(msg, key_decoder, value_decoder):
        result = []
        msg.next_frame()
        while not CodecUtil.next_frame_is_data_structure_end_frame(msg):
            key = key_decoder(msg)
            value = value_decoder(msg)
            result.append((key, value))

        msg.next_frame()
        return result

    @staticmethod
    def decode_nullable(msg, key_decoder, value_decoder):
        if CodecUtil.next_frame_is_null_frame(msg):
            return None
        else:
            return EntryListCodec.decode(msg, key_decoder, value_decoder)


_UUID_LONG_ENTRY_SIZE_IN_BYTES = UUID_SIZE_IN_BYTES + LONG_SIZE_IN_BYTES


class EntryListUUIDLongCodec(object):
    @staticmethod
    def encode(buf, entries, is_final=False):
        n = len(entries)
        size = SIZE_OF_FRAME_LENGTH_AND_FLAGS + n * _UUID_LONG_ENTRY_SIZE_IN_BYTES
        b = bytearray(size)
        LE_INT.pack_into(b, 0, size)
        if is_final:
            LE_UINT16.pack_into(b, INT_SIZE_IN_BYTES, _IS_FINAL_FLAG)
        for i in range(n):
            key, value = entries[i]
            o = SIZE_OF_FRAME_LENGTH_AND_FLAGS + i * _UUID_LONG_ENTRY_SIZE_IN_BYTES
            FixSizedTypesCodec.encode_uuid(b, o, key)
            FixSizedTypesCodec.encode_long(b, o + UUID_SIZE_IN_BYTES, value)
        buf.extend(b)

    @staticmethod
    def decode(msg):
        b = msg.next_frame().buf
        n = len(b) // _UUID_LONG_ENTRY_SIZE_IN_BYTES
        result = []
        for i in range(n):
            o = i * _UUID_LONG_ENTRY_SIZE_IN_BYTES
            key = FixSizedTypesCodec.decode_uuid(b, o)
            value = FixSizedTypesCodec.decode_long(b, o + UUID_SIZE_IN_BYTES)
            result.append((key, value))
        return result


class EntryListUUIDListIntegerCodec(object):
    @staticmethod
    def encode(buf, entries, is_final=False):
        keys = []
        buf.extend(BEGIN_FRAME_BUF)
        for key, value in entries:
            keys.append(key)
            ListIntegerCodec.encode(buf, value)
        buf.extend(END_FRAME_BUF)
        ListUUIDCodec.encode(buf, keys, is_final)

    @staticmethod
    def decode(msg):
        values = ListMultiFrameCodec.decode(msg, ListIntegerCodec.decode)
        keys = ListUUIDCodec.decode(msg)
        result = []
        n = len(keys)
        for i in range(n):
            result.append((keys[i], values[i]))
        return result


class FixSizedTypesCodec(object):
    @staticmethod
    def encode_int(buf, offset, value):
        LE_INT.pack_into(buf, offset, value)

    @staticmethod
    def decode_int(buf, offset):
        return LE_INT.unpack_from(buf, offset)[0]

    @staticmethod
    def encode_long(buf, offset, value):
        LE_LONG.pack_into(buf, offset, value)

    @staticmethod
    def decode_long(buf, offset):
        return LE_LONG.unpack_from(buf, offset)[0]

    @staticmethod
    def encode_boolean(buf, offset, value):
        if value:
            LE_INT8.pack_into(buf, offset, 1)
        else:
            LE_INT8.pack_into(buf, offset, 0)

    @staticmethod
    def decode_boolean(buf, offset):
        return LE_INT8.unpack_from(buf, offset)[0] == 1

    @staticmethod
    def encode_byte(buf, offset, value):
        LE_INT8.pack_into(buf, offset, value)

    @staticmethod
    def decode_byte(buf, offset):
        return LE_INT8.unpack_from(buf, offset)[0]

    @staticmethod
    def encode_uuid(buf, offset, value):
        is_null = value is None
        FixSizedTypesCodec.encode_boolean(buf, offset, is_null)
        if is_null:
            return

        o = offset + BOOLEAN_SIZE_IN_BYTES
        LE_ULONG.pack_into(buf, o, value.int >> UUID_MSB_SHIFT)
        LE_ULONG.pack_into(buf, o + LONG_SIZE_IN_BYTES, value.int & UUID_LSB_MASK)

    @staticmethod
    def decode_uuid(buf, offset):
        is_null = FixSizedTypesCodec.decode_boolean(buf, offset)
        if is_null:
            return None

        msb_offset = offset + BOOLEAN_SIZE_IN_BYTES
        lsb_offset = msb_offset + LONG_SIZE_IN_BYTES
        b = (
            buf[lsb_offset - 1 : msb_offset - 1 : -1]
            + buf[lsb_offset + LONG_SIZE_IN_BYTES - 1 : lsb_offset - 1 : -1]
        )
        return uuid.UUID(bytes=bytes(b))

    @staticmethod
    def decode_short(buf, offset):
        return LE_INT16.unpack_from(buf, offset)[0]

    @staticmethod
    def decode_float(buf, offset):
        return LE_FLOAT.unpack_from(buf, offset)[0]

    @staticmethod
    def decode_double(buf, offset):
        return LE_DOUBLE.unpack_from(buf, offset)[0]

    @staticmethod
    def decode_local_date(buf, offset):
        year = FixSizedTypesCodec.decode_short(buf, offset)
        month = FixSizedTypesCodec.decode_byte(buf, offset + SHORT_SIZE_IN_BYTES)
        day = FixSizedTypesCodec.decode_byte(buf, offset + SHORT_SIZE_IN_BYTES + BYTE_SIZE_IN_BYTES)

        return date(year, month, day)

    @staticmethod
    def decode_local_time(buf, offset):
        hour = FixSizedTypesCodec.decode_byte(buf, offset)
        minute = FixSizedTypesCodec.decode_byte(buf, offset + BYTE_SIZE_IN_BYTES)
        second = FixSizedTypesCodec.decode_byte(buf, offset + BYTE_SIZE_IN_BYTES * 2)
        nano = FixSizedTypesCodec.decode_int(buf, offset + BYTE_SIZE_IN_BYTES * 3)

        return time(hour, minute, second, int(nano / 1000.0))

    @staticmethod
    def decode_local_date_time(buf, offset):
        date_value = FixSizedTypesCodec.decode_local_date(buf, offset)
        time_value = FixSizedTypesCodec.decode_local_time(buf, offset + _LOCAL_DATE_SIZE_IN_BYTES)

        return datetime.combine(date_value, time_value)

    @staticmethod
    def decode_offset_date_time(buf, offset):
        datetime_value = FixSizedTypesCodec.decode_local_date_time(buf, offset)
        offset_seconds = FixSizedTypesCodec.decode_int(buf, offset + _LOCAL_DATE_TIME_SIZE_IN_BYTES)

        return datetime_value.replace(tzinfo=timezone(timedelta(seconds=offset_seconds)))


class ListIntegerCodec(object):
    @staticmethod
    def encode(buf, arr, is_final=False):
        n = len(arr)
        size = SIZE_OF_FRAME_LENGTH_AND_FLAGS + n * INT_SIZE_IN_BYTES
        b = bytearray(size)
        LE_INT.pack_into(b, 0, size)
        if is_final:
            LE_UINT16.pack_into(b, INT_SIZE_IN_BYTES, _IS_FINAL_FLAG)
        for i in range(n):
            FixSizedTypesCodec.encode_int(
                b, SIZE_OF_FRAME_LENGTH_AND_FLAGS + i * INT_SIZE_IN_BYTES, arr[i]
            )
        buf.extend(b)

    @staticmethod
    def decode(msg):
        b = msg.next_frame().buf
        n = len(b) // INT_SIZE_IN_BYTES
        result = []
        for i in range(n):
            result.append(FixSizedTypesCodec.decode_int(b, i * INT_SIZE_IN_BYTES))
        return result


class ListLongCodec(object):
    @staticmethod
    def encode(buf, arr, is_final=False):
        n = len(arr)
        size = SIZE_OF_FRAME_LENGTH_AND_FLAGS + n * LONG_SIZE_IN_BYTES
        b = bytearray(size)
        LE_INT.pack_into(b, 0, size)
        if is_final:
            LE_UINT16.pack_into(b, INT_SIZE_IN_BYTES, _IS_FINAL_FLAG)
        for i in range(n):
            FixSizedTypesCodec.encode_long(
                b, SIZE_OF_FRAME_LENGTH_AND_FLAGS + i * LONG_SIZE_IN_BYTES, arr[i]
            )
        buf.extend(b)

    @staticmethod
    def decode(msg):
        b = msg.next_frame().buf
        n = len(b) // LONG_SIZE_IN_BYTES
        result = []
        for i in range(n):
            result.append(FixSizedTypesCodec.decode_long(b, i * LONG_SIZE_IN_BYTES))
        return result


class ListMultiFrameCodec(object):
    @staticmethod
    def encode(buf, arr, encoder, is_final=False):
        buf.extend(BEGIN_FRAME_BUF)
        for item in arr:
            encoder(buf, item)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def encode_contains_nullable(buf, arr, encoder, is_final=False):
        buf.extend(BEGIN_FRAME_BUF)
        for item in arr:
            if item is None:
                buf.extend(NULL_FRAME_BUF)
            else:
                encoder(buf, item)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def encode_nullable(buf, arr, encoder, is_final=False):
        if arr is None:
            if is_final:
                buf.extend(NULL_FINAL_FRAME_BUF)
            else:
                buf.extend(NULL_FRAME_BUF)
        else:
            ListMultiFrameCodec.encode(buf, arr, encoder, is_final)

    @staticmethod
    def decode(msg, decoder):
        result = []
        msg.next_frame()
        while not CodecUtil.next_frame_is_data_structure_end_frame(msg):
            result.append(decoder(msg))

        msg.next_frame()
        return result

    @staticmethod
    def decode_contains_nullable(msg, decoder):
        result = []
        msg.next_frame()
        while not CodecUtil.next_frame_is_data_structure_end_frame(msg):
            if CodecUtil.next_frame_is_null_frame(msg):
                result.append(None)
            else:
                result.append(decoder(msg))

        msg.next_frame()
        return result

    @staticmethod
    def decode_nullable(msg, decoder):
        if CodecUtil.next_frame_is_null_frame(msg):
            return None
        else:
            return ListMultiFrameCodec.decode(msg, decoder)


class ListUUIDCodec(object):
    @staticmethod
    def encode(buf, arr, is_final=False):
        n = len(arr)
        size = SIZE_OF_FRAME_LENGTH_AND_FLAGS + n * UUID_SIZE_IN_BYTES
        b = bytearray(size)
        LE_INT.pack_into(b, 0, size)
        if is_final:
            LE_UINT16.pack_into(b, INT_SIZE_IN_BYTES, _IS_FINAL_FLAG)
        for i in range(n):
            FixSizedTypesCodec.encode_uuid(
                b, SIZE_OF_FRAME_LENGTH_AND_FLAGS + i * UUID_SIZE_IN_BYTES, arr[i]
            )
        buf.extend(b)

    @staticmethod
    def decode(msg):
        b = msg.next_frame().buf
        n = len(b) // UUID_SIZE_IN_BYTES
        result = []
        for i in range(n):
            result.append(FixSizedTypesCodec.decode_uuid(b, i * UUID_SIZE_IN_BYTES))
        return result


class LongArrayCodec(object):
    @staticmethod
    def encode(buf, arr, is_final=False):
        n = len(arr)
        size = SIZE_OF_FRAME_LENGTH_AND_FLAGS + n * LONG_SIZE_IN_BYTES
        b = bytearray(size)
        LE_INT.pack_into(b, 0, size)
        if is_final:
            LE_UINT16.pack_into(b, INT_SIZE_IN_BYTES, _IS_FINAL_FLAG)
        for i in range(n):
            FixSizedTypesCodec.encode_long(
                b, SIZE_OF_FRAME_LENGTH_AND_FLAGS + i * LONG_SIZE_IN_BYTES, arr[i]
            )
        buf.extend(b)

    @staticmethod
    def decode(msg):
        b = msg.next_frame().buf
        n = len(b) // LONG_SIZE_IN_BYTES
        result = []
        for i in range(n):
            result.append(FixSizedTypesCodec.decode_long(b, i * LONG_SIZE_IN_BYTES))
        return result


class MapCodec(object):
    @staticmethod
    def encode(buf, m, key_encoder, value_encoder, is_final=False):
        buf.extend(BEGIN_FRAME_BUF)
        for key, value in six.iteritems(m):
            key_encoder(buf, key)
            value_encoder(buf, value)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def encode_nullable(buf, m, key_encoder, value_encoder, is_final=False):
        if m is None:
            if is_final:
                buf.extend(NULL_FINAL_FRAME_BUF)
            else:
                buf.extend(NULL_FRAME_BUF)
        else:
            MapCodec.encode(buf, m, key_encoder, value_encoder, is_final)

    @staticmethod
    def decode(msg, key_decoder, value_decoder):
        result = dict()
        msg.next_frame()
        while not CodecUtil.next_frame_is_data_structure_end_frame(msg):
            key = key_decoder(msg)
            value = value_decoder(msg)
            result[key] = value

        msg.next_frame()
        return result

    @staticmethod
    def decode_nullable(msg, key_decoder, value_decoder):
        if CodecUtil.next_frame_is_null_frame(msg):
            return None
        else:
            return MapCodec.decode(msg, key_decoder, value_decoder)


class StringCodec(object):
    @staticmethod
    def encode(buf, value, is_final=False):
        value_bytes = value.encode("utf-8")
        header = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
        LE_INT.pack_into(header, 0, SIZE_OF_FRAME_LENGTH_AND_FLAGS + len(value_bytes))
        if is_final:
            LE_UINT16.pack_into(header, INT_SIZE_IN_BYTES, _IS_FINAL_FLAG)
        buf.extend(header)
        buf.extend(value_bytes)

    @staticmethod
    def decode(msg):
        return msg.next_frame().buf.decode("utf-8")


class ListCNFixedSizeCodec(object):
    _TYPE_NULL_ONLY = 1
    _TYPE_NOT_NULL_ONLY = 2
    _TYPE_MIXED = 3

    _ITEMS_PER_BITMASK = 8

    _HEADER_SIZE = BYTE_SIZE_IN_BYTES + INT_SIZE_IN_BYTES

    @staticmethod
    def decode(msg, item_size, decoder):
        frame = msg.next_frame()
        type = FixSizedTypesCodec.decode_byte(frame.buf, 0)
        count = FixSizedTypesCodec.decode_int(frame.buf, 1)

        if type == ListCNFixedSizeCodec._TYPE_NULL_ONLY:
            return [None] * count
        elif type == ListCNFixedSizeCodec._TYPE_NOT_NULL_ONLY:
            header_size = ListCNFixedSizeCodec._HEADER_SIZE
            return [decoder(frame.buf, header_size + i * item_size) for i in range(count)]
        else:
            response = [None] * count
            position = ListCNFixedSizeCodec._HEADER_SIZE
            read_count = 0
            items_per_bitmask = ListCNFixedSizeCodec._ITEMS_PER_BITMASK

            while read_count < count:
                bitmask = FixSizedTypesCodec.decode_byte(frame.buf, position)
                position += 1

                batch_size = min(items_per_bitmask, count - read_count)
                for i in range(batch_size):
                    mask = 1 << i
                    if (bitmask & mask) == mask:
                        response[read_count] = decoder(frame.buf, position)
                        position += item_size

                    read_count += 1

            return response


class ListCNBooleanCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(
            msg, BOOLEAN_SIZE_IN_BYTES, FixSizedTypesCodec.decode_boolean
        )


class ListCNByteCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(msg, BYTE_SIZE_IN_BYTES, FixSizedTypesCodec.decode_byte)


class ListCNShortCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(
            msg, SHORT_SIZE_IN_BYTES, FixSizedTypesCodec.decode_short
        )


class ListCNIntegerCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(msg, INT_SIZE_IN_BYTES, FixSizedTypesCodec.decode_int)


class ListCNLongCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(msg, LONG_SIZE_IN_BYTES, FixSizedTypesCodec.decode_long)


class ListCNFloatCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(
            msg, FLOAT_SIZE_IN_BYTES, FixSizedTypesCodec.decode_float
        )


class ListCNDoubleCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(
            msg, DOUBLE_SIZE_IN_BYTES, FixSizedTypesCodec.decode_double
        )


class ListCNLocalDateCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(
            msg, _LOCAL_DATE_SIZE_IN_BYTES, ListCNLocalDateCodec._decode_item
        )

    @staticmethod
    def _decode_item(buf, offset):
        return FixSizedTypesCodec.decode_local_date(buf, offset).isoformat()


class ListCNLocalTimeCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(
            msg, _LOCAL_TIME_SIZE_IN_BYTES, ListCNLocalTimeCodec._decode_item
        )

    @staticmethod
    def _decode_item(buf, offset):
        return FixSizedTypesCodec.decode_local_time(buf, offset).isoformat()


class ListCNLocalDateTimeCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(
            msg, _LOCAL_DATE_TIME_SIZE_IN_BYTES, ListCNLocalDateTimeCodec._decode_item
        )

    @staticmethod
    def _decode_item(buf, offset):
        return FixSizedTypesCodec.decode_local_date_time(buf, offset).isoformat()


class ListCNOffsetDateTimeCodec(object):
    @staticmethod
    def decode(msg):
        return ListCNFixedSizeCodec.decode(
            msg, _OFFSET_DATE_TIME_SIZE_IN_BYTES, ListCNOffsetDateTimeCodec._decode_item
        )

    @staticmethod
    def _decode_item(buf, offset):
        return FixSizedTypesCodec.decode_offset_date_time(buf, offset).isoformat()


class BigDecimalCodec(object):
    @staticmethod
    def decode(msg):
        buf = msg.next_frame().buf
        size = FixSizedTypesCodec.decode_int(buf, 0)
        unscaled_value = int_from_bytes(buf[INT_SIZE_IN_BYTES : INT_SIZE_IN_BYTES + size])
        scale = FixSizedTypesCodec.decode_int(buf, INT_SIZE_IN_BYTES + size)
        sign = 0 if unscaled_value >= 0 else 1
        return str(
            Decimal((sign, tuple(int(digit) for digit in str(abs(unscaled_value))), -1 * scale))
        )


class SqlPageCodec(object):
    @staticmethod
    def decode(msg):
        from hazelcast.sql import SqlColumnType, _SqlPage

        # begin frame
        msg.next_frame()

        # read the "last" flag
        is_last = LE_INT8.unpack_from(msg.next_frame().buf, 0)[0] == 1

        # read column types
        column_type_ids = ListIntegerCodec.decode(msg)
        column_count = len(column_type_ids)

        # read columns
        columns = [None] * column_count

        for i in range(column_count):
            column_type_id = column_type_ids[i]

            if column_type_id == SqlColumnType.VARCHAR:
                columns[i] = ListMultiFrameCodec.decode_contains_nullable(msg, StringCodec.decode)
            elif column_type_id == SqlColumnType.BOOLEAN:
                columns[i] = ListCNBooleanCodec.decode(msg)
            elif column_type_id == SqlColumnType.TINYINT:
                columns[i] = ListCNByteCodec.decode(msg)
            elif column_type_id == SqlColumnType.SMALLINT:
                columns[i] = ListCNShortCodec.decode(msg)
            elif column_type_id == SqlColumnType.INTEGER:
                columns[i] = ListCNIntegerCodec.decode(msg)
            elif column_type_id == SqlColumnType.BIGINT:
                columns[i] = ListCNLongCodec.decode(msg)
            elif column_type_id == SqlColumnType.REAL:
                columns[i] = ListCNFloatCodec.decode(msg)
            elif column_type_id == SqlColumnType.DOUBLE:
                columns[i] = ListCNDoubleCodec.decode(msg)
            elif column_type_id == SqlColumnType.DATE:
                columns[i] = ListCNLocalDateCodec.decode(msg)
            elif column_type_id == SqlColumnType.TIME:
                columns[i] = ListCNLocalTimeCodec.decode(msg)
            elif column_type_id == SqlColumnType.TIMESTAMP:
                columns[i] = ListCNLocalDateTimeCodec.decode(msg)
            elif column_type_id == SqlColumnType.TIMESTAMP_WITH_TIME_ZONE:
                columns[i] = ListCNOffsetDateTimeCodec.decode(msg)
            elif column_type_id == SqlColumnType.DECIMAL:
                columns[i] = ListMultiFrameCodec.decode_contains_nullable(
                    msg, BigDecimalCodec.decode
                )
            elif column_type_id == SqlColumnType.NULL:
                frame = msg.next_frame()
                size = FixSizedTypesCodec.decode_int(frame.buf, 0)
                column = [None for _ in range(size)]
                columns[i] = column
            elif column_type_id == SqlColumnType.OBJECT:
                columns[i] = ListMultiFrameCodec.decode_contains_nullable(msg, DataCodec.decode)
            else:
                raise ValueError("Unknown type %s" % column_type_id)

        CodecUtil.fast_forward_to_end_frame(msg)

        return _SqlPage(column_type_ids, columns, is_last)



import logging
import threading
import time
from hazelcast.errors import TransactionError, IllegalStateError
from hazelcast.future import make_blocking
from hazelcast.invocation import Invocation
from hazelcast.protocol.codec import (
    transaction_create_codec,
    transaction_commit_codec,
    transaction_rollback_codec,
)
from hazelcast.proxy.transactional_list import TransactionalList
from hazelcast.proxy.transactional_map import TransactionalMap
from hazelcast.proxy.transactional_multi_map import TransactionalMultiMap
from hazelcast.proxy.transactional_queue import TransactionalQueue
from hazelcast.proxy.transactional_set import TransactionalSet
from hazelcast.util import thread_id
from hazelcast.six.moves import range

_logger = logging.getLogger(__name__)

_STATE_ACTIVE = "active"
_STATE_NOT_STARTED = "not_started"
_STATE_COMMITTED = "committed"
_STATE_ROLLED_BACK = "rolled_back"
_STATE_PARTIAL_COMMIT = "rolling_back"

TWO_PHASE = 1
"""
The two phase commit is separated in 2 parts. First it tries to execute the prepare; if there are any conflicts,
the prepare will fail. Once the prepare has succeeded, the commit (writing the changes) can be executed.

Hazelcast also provides three phase transaction by automatically copying the backlog to another member so that in case
of failure during a commit, another member can continue the commit from backup.
"""

ONE_PHASE = 2
"""
The one phase transaction executes a transaction using a single step at the end; committing the changes. There
is no prepare of the transactions, so conflicts are not detected. If there is a conflict, then when the transaction
commits the changes, some of the changes are written and others are not; leaving the system in a potentially permanent
inconsistent state.
"""

RETRY_COUNT = 20


class TransactionManager(object):
    """Manages the execution of client transactions and provides Transaction objects."""

    def __init__(self, context):
        self._context = context

    def _connect(self):
        connection_manager = self._context.connection_manager
        for count in range(0, RETRY_COUNT):
            connection = connection_manager.get_random_connection()
            if connection:
                return connection

            _logger.debug(
                "Could not get a connection for the transaction. Attempt %d of %d",
                count,
                RETRY_COUNT,
                exc_info=True,
            )
            if count + 1 == RETRY_COUNT:
                raise IllegalStateError("No active connection is found")

    def new_transaction(self, timeout, durability, transaction_type):
        """Creates a Transaction object with given timeout, durability and transaction type.

        Args:
            timeout (int): The timeout in seconds determines the maximum lifespan of a transaction.
            durability (int): The durability is the number of machines that can take over if a member fails during a
                transaction commit or rollback
            transaction_type (int): the transaction type which can be ``hazelcast.transaction.TWO_PHASE``
                or ``hazelcast.transaction.ONE_PHASE``

        Returns:
          hazelcast.transaction.Transaction: New created Transaction.
        """
        connection = self._connect()
        return Transaction(self._context, connection, timeout, durability, transaction_type)


class Transaction(object):
    """Provides transactional operations: beginning/committing transactions, but also retrieving
    transactional data-structures like the TransactionalMap.
    """

    state = _STATE_NOT_STARTED
    id = None
    start_time = None
    _locals = threading.local()
    thread_id = None

    def __init__(self, context, connection, timeout, durability, transaction_type):
        self._context = context
        self.connection = connection
        self.timeout = timeout
        self.durability = durability
        self.transaction_type = transaction_type
        self._objects = {}

    def begin(self):
        """Begins this transaction."""
        if hasattr(self._locals, "transaction_exists") and self._locals.transaction_exists:
            raise TransactionError("Nested transactions are not allowed.")
        if self.state != _STATE_NOT_STARTED:
            raise TransactionError("Transaction has already been started.")
        self._locals.transaction_exists = True
        self.start_time = time.time()
        self.thread_id = thread_id()
        try:
            request = transaction_create_codec.encode_request(
                timeout=int(self.timeout * 1000),
                durability=self.durability,
                transaction_type=self.transaction_type,
                thread_id=self.thread_id,
            )
            invocation = Invocation(
                request, connection=self.connection, response_handler=lambda m: m
            )
            invocation_service = self._context.invocation_service
            invocation_service.invoke(invocation)
            response = invocation.future.result()
            self.id = transaction_create_codec.decode_response(response)
            self.state = _STATE_ACTIVE
        except:
            self._locals.transaction_exists = False
            raise

    def commit(self):
        """Commits this transaction."""
        self._check_thread()
        if self.state != _STATE_ACTIVE:
            raise TransactionError("Transaction is not active.")
        try:
            self._check_timeout()
            request = transaction_commit_codec.encode_request(self.id, self.thread_id)
            invocation = Invocation(request, connection=self.connection)
            invocation_service = self._context.invocation_service
            invocation_service.invoke(invocation)
            invocation.future.result()
            self.state = _STATE_COMMITTED
        except:
            self.state = _STATE_PARTIAL_COMMIT
            raise
        finally:
            self._locals.transaction_exists = False

    def rollback(self):
        """Rollback of this current transaction."""
        self._check_thread()
        if self.state not in (_STATE_ACTIVE, _STATE_PARTIAL_COMMIT):
            raise TransactionError("Transaction is not active.")
        try:
            if self.state != _STATE_PARTIAL_COMMIT:
                request = transaction_rollback_codec.encode_request(self.id, self.thread_id)
                invocation = Invocation(request, connection=self.connection)
                invocation_service = self._context.invocation_service
                invocation_service.invoke(invocation)
                invocation.future.result()
            self.state = _STATE_ROLLED_BACK
        finally:
            self._locals.transaction_exists = False

    def get_list(self, name):
        """Returns the transactional list instance with the specified name.

        Args:
            name (str): The specified name.

        Returns:
            hazelcast.proxy.transactional_list.TransactionalList`: The instance of Transactional List
                with the specified name.
        """
        return self._get_or_create_object(name, TransactionalList)

    def get_map(self, name):
        """Returns the transactional map instance with the specified name.

        Args:
            name (str): The specified name.

        Returns:
            hazelcast.proxy.transactional_map.TransactionalMap: The instance of Transactional Map
                with the specified name.
        """
        return self._get_or_create_object(name, TransactionalMap)

    def get_multi_map(self, name):
        """Returns the transactional multimap instance with the specified name.

        Args:
            name (str): The specified name.

        Returns:
            hazelcast.proxy.transactional_multi_map.TransactionalMultiMap: The instance of Transactional MultiMap
                with the specified name.
        """
        return self._get_or_create_object(name, TransactionalMultiMap)

    def get_queue(self, name):
        """Returns the transactional queue instance with the specified name.

        Args:
            name (str): The specified name.

        Returns:
            hazelcast.proxy.transactional_queue.TransactionalQueue: The instance of Transactional Queue
                with the specified name.
        """
        return self._get_or_create_object(name, TransactionalQueue)

    def get_set(self, name):
        """Returns the transactional set instance with the specified name.

        Args:
            name (str): The specified name.

        Returns:
            hazelcast.proxy.transactional_set.TransactionalSet: The instance of Transactional Set
                with the specified name.
        """
        return self._get_or_create_object(name, TransactionalSet)

    def _get_or_create_object(self, name, proxy_type):
        if self.state != _STATE_ACTIVE:
            raise TransactionError("Transaction is not in active state.")
        self._check_thread()
        key = (proxy_type, name)
        try:
            return self._objects[key]
        except KeyError:
            proxy = proxy_type(name, self, self._context)
            self._objects[key] = proxy
            return make_blocking(proxy)

    def _check_thread(self):
        if not thread_id() == self.thread_id:
            raise TransactionError("Transaction cannot span multiple threads.")

    def _check_timeout(self):
        if time.time() > self.timeout + self.start_time:
            raise TransactionError("Transaction has timed out.")

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, type, value, traceback):
        if not type and not value and self.state == _STATE_ACTIVE:
            self.commit()
        elif self.state in (_STATE_PARTIAL_COMMIT, _STATE_ACTIVE):
            self.rollback()



import socket

EXCEPTION_MESSAGE_TYPE = 0


def retryable(cls):
    cls.retryable = True
    return cls


class HazelcastError(Exception):
    """General HazelcastError class."""

    def __init__(self, message=None, cause=None):
        super(HazelcastError, self).__init__(message, cause)

    def __str__(self):
        message, cause = self.args
        if cause:
            return "%s\nCaused by: %s" % (message, str(cause))
        return message or self.__class__.__name__


class ArrayIndexOutOfBoundsError(HazelcastError):
    pass


class ArrayStoreError(HazelcastError):
    pass


class AuthenticationError(HazelcastError):
    pass


class CacheNotExistsError(HazelcastError):
    pass


@retryable
class CallerNotMemberError(HazelcastError):
    pass


class CancellationError(HazelcastError):
    pass


class ClassCastError(HazelcastError):
    pass


class ClassNotFoundError(HazelcastError):
    pass


class ConcurrentModificationError(HazelcastError):
    pass


class ConfigMismatchError(HazelcastError):
    pass


class ConfigurationError(HazelcastError):
    pass


class DistributedObjectDestroyedError(HazelcastError):
    pass


class DuplicateInstanceNameError(HazelcastError):
    pass


class HazelcastEOFError(HazelcastError):
    pass


class ExecutionError(HazelcastError):
    pass


class HazelcastInstanceNotActiveError(HazelcastError):
    pass


class HazelcastOverloadError(HazelcastError):
    pass


class HazelcastSerializationError(HazelcastError):
    pass


class HazelcastIOError(HazelcastError):
    pass


class IllegalArgumentError(HazelcastError):
    pass


class IllegalAccessException(HazelcastError):
    pass


class IllegalAccessError(HazelcastError):
    pass


class IllegalMonitorStateError(HazelcastError):
    pass


class IllegalStateError(HazelcastError):
    pass


class IllegalThreadStateError(HazelcastError):
    pass


class IndexOutOfBoundsError(HazelcastError):
    pass


class HazelcastInterruptedError(HazelcastError):
    pass


class InvalidAddressError(HazelcastError):
    pass


class InvalidConfigurationError(HazelcastError):
    pass


@retryable
class MemberLeftError(HazelcastError):
    pass


class NegativeArraySizeError(HazelcastError):
    pass


class NoSuchElementError(HazelcastError):
    pass


class NotSerializableError(HazelcastError):
    pass


class NullPointerError(HazelcastError):
    pass


class OperationTimeoutError(HazelcastError):
    pass


@retryable
class PartitionMigratingError(HazelcastError):
    pass


class QueryError(HazelcastError):
    pass


class QueryResultSizeExceededError(HazelcastError):
    pass


class SplitBrainProtectionError(HazelcastError):
    pass


class ReachedMaxSizeError(HazelcastError):
    pass


class RejectedExecutionError(HazelcastError):
    pass


class ResponseAlreadySentError(HazelcastError):
    pass


@retryable
class RetryableHazelcastError(HazelcastError):
    pass


@retryable
class RetryableIOError(HazelcastError):
    pass


class HazelcastRuntimeError(HazelcastError):
    pass


class SecurityError(HazelcastError):
    pass


class SocketError(HazelcastError):
    pass


class StaleSequenceError(HazelcastError):
    pass


class TargetDisconnectedError(HazelcastError):
    pass


@retryable
class TargetNotMemberError(HazelcastError):
    pass


class HazelcastTimeoutError(HazelcastError):
    pass


class TopicOverloadError(HazelcastError):
    pass


class TransactionError(HazelcastError):
    pass


class TransactionNotActiveError(HazelcastError):
    pass


class TransactionTimedOutError(HazelcastError):
    pass


class URISyntaxError(HazelcastError):
    pass


class UTFDataFormatError(HazelcastError):
    pass


class UnsupportedOperationError(HazelcastError):
    pass


@retryable
class WrongTargetError(HazelcastError):
    pass


class XAError(HazelcastError):
    pass


class AccessControlError(HazelcastError):
    pass


class LoginError(HazelcastError):
    pass


class UnsupportedCallbackError(HazelcastError):
    pass


class NoDataMemberInClusterError(HazelcastError):
    pass


class ReplicatedMapCantBeCreatedOnLiteMemberError(HazelcastError):
    pass


class MaxMessageSizeExceededError(HazelcastError):
    pass


class WANReplicationQueueFullError(HazelcastError):
    pass


class HazelcastAssertionError(HazelcastError):
    pass


class OutOfMemoryError(HazelcastError):
    pass


class StackOverflowError(HazelcastError):
    pass


class NativeOutOfMemoryError(HazelcastError):
    pass


class ServiceNotFoundError(HazelcastError):
    pass


class StaleTaskIdError(HazelcastError):
    pass


class DuplicateTaskError(HazelcastError):
    pass


class StaleTaskError(HazelcastError):
    pass


class LocalMemberResetError(HazelcastError):
    pass


class IndeterminateOperationStateError(HazelcastError):
    pass


class NodeIdOutOfRangeError(HazelcastError):
    pass


@retryable
class TargetNotReplicaError(HazelcastError):
    pass


class MutationDisallowedError(HazelcastError):
    pass


class ConsistencyLostError(HazelcastError):
    pass


class HazelcastClientNotActiveError(HazelcastError):
    def __init__(self, message="Client is not active"):
        super(HazelcastClientNotActiveError, self).__init__(message)


class HazelcastCertificationError(HazelcastError):
    pass


class ClientOfflineError(HazelcastError):
    def __init__(self):
        super(ClientOfflineError, self).__init__("No connection found to cluster")


class ClientNotAllowedInClusterError(HazelcastError):
    pass


class VersionMismatchError(HazelcastError):
    pass


class NoSuchMethodError(HazelcastError):
    pass


class NoSuchMethodException(HazelcastError):
    pass


class NoSuchFieldError(HazelcastError):
    pass


class NoSuchFieldException(HazelcastError):
    pass


class NoClassDefFoundError(HazelcastError):
    pass


class UndefinedErrorCodeError(HazelcastError):
    pass


class SessionExpiredError(HazelcastError):
    pass


class WaitKeyCancelledError(HazelcastError):
    pass


class LockAcquireLimitReachedError(HazelcastError):
    pass


class LockOwnershipLostError(HazelcastError):
    pass


class CPGroupDestroyedError(HazelcastError):
    pass


@retryable
class CannotReplicateError(HazelcastError):
    pass


class LeaderDemotedError(HazelcastError):
    pass


class StaleAppendRequestError(HazelcastError):
    pass


class NotLeaderError(HazelcastError):
    pass


# Error Codes
_UNDEFINED = 0
_ARRAY_INDEX_OUT_OF_BOUNDS = 1
_ARRAY_STORE = 2
_AUTHENTICATION = 3
_CACHE = 4
_CACHE_LOADER = 5
_CACHE_NOT_EXISTS = 6
_CACHE_WRITER = 7
_CALLER_NOT_MEMBER = 8
_CANCELLATION = 9
_CLASS_CAST = 10
_CLASS_NOT_FOUND = 11
_CONCURRENT_MODIFICATION = 12
_CONFIG_MISMATCH = 13
_DISTRIBUTED_OBJECT_DESTROYED = 14
_EOF = 15
_ENTRY_PROCESSOR = 16
_EXECUTION = 17
_HAZELCAST = 18
_HAZELCAST_INSTANCE_NOT_ACTIVE = 19
_HAZELCAST_OVERLOAD = 20
_HAZELCAST_SERIALIZATION = 21
_IO = 22
_ILLEGAL_ARGUMENT = 23
_ILLEGAL_ACCESS_EXCEPTION = 24
_ILLEGAL_ACCESS_ERROR = 25
_ILLEGAL_MONITOR_STATE = 26
_ILLEGAL_STATE = 27
_ILLEGAL_THREAD_STATE = 28
_INDEX_OUT_OF_BOUNDS = 29
_INTERRUPTED = 30
_INVALID_ADDRESS = 31
_INVALID_CONFIGURATION = 32
_MEMBER_LEFT = 33
_NEGATIVE_ARRAY_SIZE = 34
_NO_SUCH_ELEMENT = 35
_NOT_SERIALIZABLE = 36
_NULL_POINTER = 37
_OPERATION_TIMEOUT = 38
_PARTITION_MIGRATING = 39
_QUERY = 40
_QUERY_RESULT_SIZE_EXCEEDED = 41
_SPLIT_BRAIN_PROTECTION = 42
_REACHED_MAX_SIZE = 43
_REJECTED_EXECUTION = 44
_RESPONSE_ALREADY_SENT = 45
_RETRYABLE_HAZELCAST = 46
_RETRYABLE_IO = 47
_RUNTIME = 48
_SECURITY = 49
_SOCKET = 50
_STALE_SEQUENCE = 51
_TARGET_DISCONNECTED = 52
_TARGET_NOT_MEMBER = 53
_TIMEOUT = 54
_TOPIC_OVERLOAD = 55
_TRANSACTION = 56
_TRANSACTION_NOT_ACTIVE = 57
_TRANSACTION_TIMED_OUT = 58
_URI_SYNTAX = 59
_UTF_DATA_FORMAT = 60
_UNSUPPORTED_OPERATION = 61
_WRONG_TARGET = 62
_XA = 63
_ACCESS_CONTROL = 64
_LOGIN = 65
_UNSUPPORTED_CALLBACK = 66
_NO_DATA_MEMBER = 67
_REPLICATED_MAP_CANT_BE_CREATED = 68
_MAX_MESSAGE_SIZE_EXCEEDED = 69
_WAN_REPLICATION_QUEUE_FULL = 70
_ASSERTION_ERROR = 71
_OUT_OF_MEMORY_ERROR = 72
_STACK_OVERFLOW_ERROR = 73
_NATIVE_OUT_OF_MEMORY_ERROR = 74
_SERVICE_NOT_FOUND = 75
_STALE_TASK_ID = 76
_DUPLICATE_TASK = 77
_STALE_TASK = 78
_LOCAL_MEMBER_RESET = 79
_INDETERMINATE_OPERATION_STATE = 80
_FLAKE_ID_NODE_ID_OUT_OF_RANGE_EXCEPTION = 81
_TARGET_NOT_REPLICA_EXCEPTION = 82
_MUTATION_DISALLOWED_EXCEPTION = 83
_CONSISTENCY_LOST_EXCEPTION = 84
_SESSION_EXPIRED_EXCEPTION = 85
_WAIT_KEY_CANCELLED_EXCEPTION = 86
_LOCK_ACQUIRE_LIMIT_REACHED_EXCEPTION = 87
_LOCK_OWNERSHIP_LOST_EXCEPTION = 88
_CP_GROUP_DESTROYED_EXCEPTION = 89
_CANNOT_REPLICATE_EXCEPTION = 90
_LEADER_DEMOTED_EXCEPTION = 91
_STALE_APPEND_REQUEST_EXCEPTION = 92
_NOT_LEADER_EXCEPTION = 93
_VERSION_MISMATCH_EXCEPTION = 94
_NO_SUCH_METHOD_ERROR = 95
_NO_SUCH_METHOD_EXCEPTION = 96
_NO_SUCH_FIELD_ERROR = 97
_NO_SUCH_FIELD_EXCEPTION = 98
_NO_CLASS_DEF_FOUND_ERROR = 99

_ERROR_CODE_TO_ERROR = {
    _ARRAY_INDEX_OUT_OF_BOUNDS: ArrayIndexOutOfBoundsError,
    _ARRAY_STORE: ArrayStoreError,
    _AUTHENTICATION: AuthenticationError,
    _CACHE_NOT_EXISTS: CacheNotExistsError,
    _CALLER_NOT_MEMBER: CallerNotMemberError,
    _CANCELLATION: CancellationError,
    _CLASS_CAST: ClassCastError,
    _CLASS_NOT_FOUND: ClassNotFoundError,
    _CONCURRENT_MODIFICATION: ConcurrentModificationError,
    _CONFIG_MISMATCH: ConfigMismatchError,
    _DISTRIBUTED_OBJECT_DESTROYED: DistributedObjectDestroyedError,
    _EOF: HazelcastEOFError,
    _EXECUTION: ExecutionError,
    _HAZELCAST: HazelcastError,
    _HAZELCAST_INSTANCE_NOT_ACTIVE: HazelcastInstanceNotActiveError,
    _HAZELCAST_OVERLOAD: HazelcastOverloadError,
    _HAZELCAST_SERIALIZATION: HazelcastSerializationError,
    _IO: HazelcastIOError,
    _ILLEGAL_ARGUMENT: IllegalArgumentError,
    _ILLEGAL_ACCESS_EXCEPTION: IllegalAccessException,
    _ILLEGAL_ACCESS_ERROR: IllegalAccessError,
    _ILLEGAL_MONITOR_STATE: IllegalMonitorStateError,
    _ILLEGAL_STATE: IllegalStateError,
    _ILLEGAL_THREAD_STATE: IllegalThreadStateError,
    _INDEX_OUT_OF_BOUNDS: IndexOutOfBoundsError,
    _INTERRUPTED: HazelcastInterruptedError,
    _INVALID_ADDRESS: InvalidAddressError,
    _INVALID_CONFIGURATION: InvalidConfigurationError,
    _MEMBER_LEFT: MemberLeftError,
    _NEGATIVE_ARRAY_SIZE: NegativeArraySizeError,
    _NO_SUCH_ELEMENT: NoSuchElementError,
    _NOT_SERIALIZABLE: NotSerializableError,
    _NULL_POINTER: NullPointerError,
    _OPERATION_TIMEOUT: OperationTimeoutError,
    _PARTITION_MIGRATING: PartitionMigratingError,
    _QUERY: QueryError,
    _QUERY_RESULT_SIZE_EXCEEDED: QueryResultSizeExceededError,
    _SPLIT_BRAIN_PROTECTION: SplitBrainProtectionError,
    _REACHED_MAX_SIZE: ReachedMaxSizeError,
    _REJECTED_EXECUTION: RejectedExecutionError,
    _RESPONSE_ALREADY_SENT: ResponseAlreadySentError,
    _RETRYABLE_HAZELCAST: RetryableHazelcastError,
    _RETRYABLE_IO: RetryableIOError,
    _RUNTIME: HazelcastRuntimeError,
    _SECURITY: SecurityError,
    _SOCKET: socket.error,
    _STALE_SEQUENCE: StaleSequenceError,
    _TARGET_DISCONNECTED: TargetDisconnectedError,
    _TARGET_NOT_MEMBER: TargetNotMemberError,
    _TIMEOUT: HazelcastTimeoutError,
    _TOPIC_OVERLOAD: TopicOverloadError,
    _TRANSACTION: TransactionError,
    _TRANSACTION_NOT_ACTIVE: TransactionNotActiveError,
    _TRANSACTION_TIMED_OUT: TransactionTimedOutError,
    _URI_SYNTAX: URISyntaxError,
    _UTF_DATA_FORMAT: UTFDataFormatError,
    _UNSUPPORTED_OPERATION: UnsupportedOperationError,
    _WRONG_TARGET: WrongTargetError,
    _XA: XAError,
    _ACCESS_CONTROL: AccessControlError,
    _LOGIN: LoginError,
    _UNSUPPORTED_CALLBACK: UnsupportedCallbackError,
    _NO_DATA_MEMBER: NoDataMemberInClusterError,
    _REPLICATED_MAP_CANT_BE_CREATED: ReplicatedMapCantBeCreatedOnLiteMemberError,
    _MAX_MESSAGE_SIZE_EXCEEDED: MaxMessageSizeExceededError,
    _WAN_REPLICATION_QUEUE_FULL: WANReplicationQueueFullError,
    _ASSERTION_ERROR: HazelcastAssertionError,
    _OUT_OF_MEMORY_ERROR: OutOfMemoryError,
    _STACK_OVERFLOW_ERROR: StackOverflowError,
    _NATIVE_OUT_OF_MEMORY_ERROR: NativeOutOfMemoryError,
    _SERVICE_NOT_FOUND: ServiceNotFoundError,
    _STALE_TASK_ID: StaleTaskIdError,
    _DUPLICATE_TASK: DuplicateTaskError,
    _STALE_TASK: StaleTaskError,
    _LOCAL_MEMBER_RESET: LocalMemberResetError,
    _INDETERMINATE_OPERATION_STATE: IndeterminateOperationStateError,
    _FLAKE_ID_NODE_ID_OUT_OF_RANGE_EXCEPTION: NodeIdOutOfRangeError,
    _TARGET_NOT_REPLICA_EXCEPTION: TargetNotReplicaError,
    _MUTATION_DISALLOWED_EXCEPTION: MutationDisallowedError,
    _CONSISTENCY_LOST_EXCEPTION: ConsistencyLostError,
    _SESSION_EXPIRED_EXCEPTION: SessionExpiredError,
    _WAIT_KEY_CANCELLED_EXCEPTION: WaitKeyCancelledError,
    _LOCK_ACQUIRE_LIMIT_REACHED_EXCEPTION: LockAcquireLimitReachedError,
    _LOCK_OWNERSHIP_LOST_EXCEPTION: LockOwnershipLostError,
    _CP_GROUP_DESTROYED_EXCEPTION: CPGroupDestroyedError,
    _CANNOT_REPLICATE_EXCEPTION: CannotReplicateError,
    _LEADER_DEMOTED_EXCEPTION: LeaderDemotedError,
    _STALE_APPEND_REQUEST_EXCEPTION: StaleAppendRequestError,
    _NOT_LEADER_EXCEPTION: NotLeaderError,
    _VERSION_MISMATCH_EXCEPTION: VersionMismatchError,
    _NO_SUCH_METHOD_ERROR: NoSuchMethodError,
    _NO_SUCH_METHOD_EXCEPTION: NoSuchMethodException,
    _NO_SUCH_FIELD_ERROR: NoSuchFieldError,
    _NO_SUCH_FIELD_EXCEPTION: NoSuchFieldException,
    _NO_CLASS_DEF_FOUND_ERROR: NoClassDefFoundError,
}

from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.codec.custom.error_holder_codec import ErrorHolderCodec


class _ErrorsCodec(object):
    @staticmethod
    def decode(msg):
        msg.next_frame()
        return ListMultiFrameCodec.decode(msg, ErrorHolderCodec.decode)


def create_error_from_message(error_message):
    error_holders = _ErrorsCodec.decode(error_message)
    return _create_error(error_holders, 0)


def _create_error(error_holders, idx):
    if idx == len(error_holders):
        return None

    error_holder = error_holders[idx]
    error_class = _ERROR_CODE_TO_ERROR.get(error_holder.error_code, None)

    stack_trace = "\n".join(
        [
            "\tat %s.%s(%s:%s)" % (x.class_name, x.method_name, x.file_name, x.line_number)
            for x in error_holder.stack_trace_elements
        ]
    )
    message = "Exception from server: %s: %s\n %s" % (
        error_holder.class_name,
        error_holder.message,
        stack_trace,
    )
    if error_class:
        return error_class(message, _create_error(error_holders, idx + 1))
    else:
        return UndefinedErrorCodeError(message, error_holder.class_name)


def is_retryable_error(error):
    return hasattr(error, "retryable")



from setuptools import setup, find_packages
from codecs import open
from os import path
from hazelcast import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()


stats_requirements = [
    "psutil",
]

extras = {
    "stats": stats_requirements,
}

setup(
    name="hazelcast-python-client",
    version=__version__,
    description="Hazelcast Python Client",
    long_description=long_description,
    url="https://github.com/hazelcast/hazelcast-python-client",
    author="Hazelcast Inc. Developers",
    author_email="hazelcast@googlegroups.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    license="Apache 2.0",
    keywords="hazelcast,hazelcast client,In-Memory Data Grid,Distributed Computing",
    packages=find_packages(
        exclude=["benchmarks", "examples", "examples.*", "docs", "docs.*", "tests", "tests.*"]
    ),
    package_dir={"hazelcast": "hazelcast"},
    install_requires=[],
    extras_require=extras,
    tests_require=["thrift", "nose", "coverage", "psutil", "mock", "parameterized"],
)



"""
Serialization type ids
"""
# Serialization Constants
CONSTANT_TYPE_NULL = 0
CONSTANT_TYPE_PORTABLE = -1
CONSTANT_TYPE_DATA_SERIALIZABLE = -2
CONSTANT_TYPE_BYTE = -3
CONSTANT_TYPE_BOOLEAN = -4
CONSTANT_TYPE_CHAR = -5
CONSTANT_TYPE_SHORT = -6
CONSTANT_TYPE_INTEGER = -7
CONSTANT_TYPE_LONG = -8
CONSTANT_TYPE_FLOAT = -9
CONSTANT_TYPE_DOUBLE = -10
CONSTANT_TYPE_STRING = -11
CONSTANT_TYPE_BYTE_ARRAY = -12
CONSTANT_TYPE_BOOLEAN_ARRAY = -13
CONSTANT_TYPE_CHAR_ARRAY = -14
CONSTANT_TYPE_SHORT_ARRAY = -15
CONSTANT_TYPE_INTEGER_ARRAY = -16
CONSTANT_TYPE_LONG_ARRAY = -17
CONSTANT_TYPE_FLOAT_ARRAY = -18
CONSTANT_TYPE_DOUBLE_ARRAY = -19
CONSTANT_TYPE_STRING_ARRAY = -20
CONSTANT_TYPE_UUID = -21

# ------------------------------------------------------------
# DEFAULT SERIALIZERS

JAVA_DEFAULT_TYPE_CLASS = -24
JAVA_DEFAULT_TYPE_DATE = -25
JAVA_DEFAULT_TYPE_BIG_INTEGER = -26
JAVA_DEFAULT_TYPE_ARRAY = -28
JAVA_DEFAULT_TYPE_ARRAY_LIST = -29
JAVA_DEFAULT_TYPE_LINKED_LIST = -30
JAVASCRIPT_JSON_SERIALIZATION_TYPE = -130

# ------------------------------------------------------------
# JAVA SERIALIZATION

JAVA_DEFAULT_TYPE_SERIALIZABLE = -100
PYTHON_TYPE_PICKLE = -120



import logging
import sys
import threading
from uuid import uuid4

from hazelcast import six
from hazelcast.errors import HazelcastError, HazelcastClientNotActiveError, TargetDisconnectedError
from hazelcast.future import combine_futures, ImmediateFuture
from hazelcast.invocation import Invocation
from hazelcast.protocol.codec import client_add_cluster_view_listener_codec
from hazelcast.util import check_not_none

_logger = logging.getLogger(__name__)


class _ListenerRegistration(object):
    __slots__ = (
        "registration_request",
        "decode_register_response",
        "encode_deregister_request",
        "handler",
        "connection_registrations",
    )

    def __init__(
        self, registration_request, decode_register_response, encode_deregister_request, handler
    ):
        self.registration_request = registration_request
        self.decode_register_response = decode_register_response
        self.encode_deregister_request = encode_deregister_request
        self.handler = handler
        self.connection_registrations = {}  # Dict of Connection, EventRegistration


class _EventRegistration(object):
    __slots__ = ("server_registration_id", "correlation_id")

    def __init__(self, server_registration_id, correlation_id):
        self.server_registration_id = server_registration_id
        self.correlation_id = correlation_id


class ListenerService(object):
    def __init__(self, client, config, connection_manager, invocation_service):
        self._client = client
        self._connection_manager = connection_manager
        self._invocation_service = invocation_service
        self._is_smart = config.smart_routing
        self._active_registrations = {}  # Dict of user_registration_id, ListenerRegistration
        self._registration_lock = threading.RLock()
        self._event_handlers = {}

    def start(self):
        self._connection_manager.add_listener(self._connection_added, self._connection_removed)

    def register_listener(
        self, registration_request, decode_register_response, encode_deregister_request, handler
    ):
        with self._registration_lock:
            registration_id = str(uuid4())
            registration = _ListenerRegistration(
                registration_request, decode_register_response, encode_deregister_request, handler
            )
            self._active_registrations[registration_id] = registration

            futures = []
            for connection in list(six.itervalues(self._connection_manager.active_connections)):
                future = self._register_on_connection(registration_id, registration, connection)
                futures.append(future)

            def handler(f):
                try:
                    f.result()
                    return registration_id
                except:
                    self.deregister_listener(registration_id)
                    raise HazelcastError("Listener cannot be added")

            return combine_futures(futures).continue_with(handler)

    def deregister_listener(self, user_registration_id):
        check_not_none(user_registration_id, "None user_registration_id is not allowed!")

        with self._registration_lock:
            listener_registration = self._active_registrations.pop(user_registration_id, None)
            if not listener_registration:
                return ImmediateFuture(False)

            for connection, event_registration in six.iteritems(
                listener_registration.connection_registrations
            ):
                # Remove local handler
                self.remove_event_handler(event_registration.correlation_id)
                # The rest is for deleting the remote registration
                server_registration_id = event_registration.server_registration_id
                deregister_request = listener_registration.encode_deregister_request(
                    server_registration_id
                )
                if deregister_request is None:
                    # None means no remote registration (e.g. for backup acks)
                    continue

                invocation = Invocation(
                    deregister_request, connection=connection, timeout=sys.maxsize, urgent=True
                )
                self._invocation_service.invoke(invocation)

                def handler(f, connection=connection):
                    e = f.exception()
                    if e:
                        if isinstance(
                            e, (HazelcastClientNotActiveError, IOError, TargetDisconnectedError)
                        ):
                            return

                        _logger.warning(
                            "Deregistration of listener with ID %s has failed for address %s",
                            user_registration_id,
                            connection.remote_address,
                        )

                invocation.future.add_done_callback(handler)

            listener_registration.connection_registrations.clear()
            return ImmediateFuture(True)

    def handle_client_message(self, message, correlation_id):
        handler = self._event_handlers.get(correlation_id, None)
        if handler:
            handler(message)
        else:
            _logger.debug("Got event message with unknown correlation id: %s", message)

    def add_event_handler(self, correlation_id, event_handler):
        self._event_handlers[correlation_id] = event_handler

    def remove_event_handler(self, correlation_id):
        self._event_handlers.pop(correlation_id, None)

    def _register_on_connection(self, user_registration_id, listener_registration, connection):
        registration_map = listener_registration.connection_registrations

        if connection in registration_map:
            return

        registration_request = listener_registration.registration_request.copy()
        invocation = Invocation(
            registration_request,
            connection=connection,
            event_handler=listener_registration.handler,
            response_handler=lambda m: m,
            urgent=True,
        )
        self._invocation_service.invoke(invocation)

        def callback(f):
            try:
                response = f.result()
                server_registration_id = listener_registration.decode_register_response(response)
                correlation_id = registration_request.get_correlation_id()
                registration = _EventRegistration(server_registration_id, correlation_id)
                registration_map[connection] = registration
            except Exception as e:
                if connection.live:
                    _logger.exception(
                        "Listener %s can not be added to a new connection: %s",
                        user_registration_id,
                        connection,
                    )
                raise e

        return invocation.future.continue_with(callback)

    def _connection_added(self, connection):
        with self._registration_lock:
            for user_reg_id, listener_registration in six.iteritems(self._active_registrations):
                self._register_on_connection(user_reg_id, listener_registration, connection)

    def _connection_removed(self, connection):
        with self._registration_lock:
            for listener_registration in six.itervalues(self._active_registrations):
                event_registration = listener_registration.connection_registrations.pop(
                    connection, None
                )
                if event_registration:
                    self.remove_event_handler(event_registration.correlation_id)


class ClusterViewListenerService(object):
    def __init__(
        self, client, connection_manager, partition_service, cluster_service, invocation_service
    ):
        self._client = client
        self._partition_service = partition_service
        self._connection_manager = connection_manager
        self._cluster_service = cluster_service
        self._invocation_service = invocation_service
        self._listener_added_connection = None

    def start(self):
        self._connection_manager.add_listener(self._connection_added, self._connection_removed)

    def _connection_added(self, connection):
        self._try_register(connection)

    def _connection_removed(self, connection):
        self._try_register_to_random_connection(connection)

    def _try_register_to_random_connection(self, old_connection):
        if self._listener_added_connection is not old_connection:
            return
        self._listener_added_connection = None
        new_connection = self._connection_manager.get_random_connection()
        if new_connection:
            self._try_register(new_connection)

    def _try_register(self, connection):
        if self._listener_added_connection:
            return

        self._listener_added_connection = connection
        request = client_add_cluster_view_listener_codec.encode_request()
        invocation = Invocation(
            request, connection=connection, event_handler=self._handler(connection), urgent=True
        )
        self._cluster_service.clear_member_list_version()
        self._invocation_service.invoke(invocation)

        def callback(f):
            try:
                f.result()
            except:
                self._try_register_to_random_connection(connection)

        invocation.future.add_done_callback(callback)

    def _handler(self, connection):
        def handle_partitions_view_event(version, partitions):
            self._partition_service.handle_partitions_view_event(connection, partitions, version)

        def handle_members_view_event(member_list_version, member_infos):
            self._cluster_service.handle_members_view_event(member_list_version, member_infos)

        def inner(message):
            client_add_cluster_view_listener_codec.handle(
                message, handle_members_view_event, handle_partitions_view_event
            )

        return inner



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer

# hex: 0x150100
_REQUEST_MESSAGE_TYPE = 1376512
# hex: 0x150101
_RESPONSE_MESSAGE_TYPE = 1376513

_REQUEST_TRANSACTION_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_THREAD_ID_OFFSET = _REQUEST_TRANSACTION_ID_OFFSET + UUID_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_THREAD_ID_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(transaction_id, thread_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE, True)
    FixSizedTypesCodec.encode_uuid(buf, _REQUEST_TRANSACTION_ID_OFFSET, transaction_id)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_THREAD_ID_OFFSET, thread_id)
    return OutboundMessage(buf, False)



import logging
import uuid

from hazelcast import six, __version__

_logger = logging.getLogger(__name__)


class LifecycleState(object):
    """Lifecycle states."""

    STARTING = "STARTING"
    """
    The client is starting.
    """

    STARTED = "STARTED"
    """
    The client has started.
    """

    CONNECTED = "CONNECTED"
    """
    The client connected to a member.
    """

    SHUTTING_DOWN = "SHUTTING_DOWN"
    """
    The client is shutting down.
    """

    DISCONNECTED = "DISCONNECTED"
    """
    The client disconnected from a member.
    """

    SHUTDOWN = "SHUTDOWN"
    """
    The client has shutdown.
    """


class LifecycleService(object):
    """
    Lifecycle service for the Hazelcast client. Allows to determine
    state of the client and add or remove lifecycle listeners.
    """

    def __init__(self, internal_lifecycle_service):
        self._service = internal_lifecycle_service

    def is_running(self):
        """
        Checks whether or not the instance is running.

        Returns:
            bool: ``True`` if the client is active and running, ``False`` otherwise.
        """
        return self._service.running

    def add_listener(self, on_state_change):
        """
        Adds a listener to listen for lifecycle events.

        Args:
            on_state_change (function): Function to be called when lifecycle state is changed.

        Returns:
            str: Registration id of the listener
        """
        return self._service.add_listener(on_state_change)

    def remove_listener(self, registration_id):
        """
        Removes a lifecycle listener.

        Args:
            registration_id (str): The id of the listener to be removed.

        Returns:
            bool: ``True`` if the listener is removed successfully, ``False`` otherwise.
        """
        self._service.remove_listener(registration_id)


class _InternalLifecycleService(object):
    def __init__(self, config):
        self.running = False
        self._listeners = {}

        lifecycle_listeners = config.lifecycle_listeners
        if lifecycle_listeners:
            for listener in lifecycle_listeners:
                self.add_listener(listener)

    def start(self):
        if self.running:
            return

        self.fire_lifecycle_event(LifecycleState.STARTING)
        self.running = True
        self.fire_lifecycle_event(LifecycleState.STARTED)

    def shutdown(self):
        self.running = False

    def add_listener(self, on_state_change):
        listener_id = str(uuid.uuid4())
        self._listeners[listener_id] = on_state_change
        return listener_id

    def remove_listener(self, registration_id):
        try:
            self._listeners.pop(registration_id)
            return True
        except KeyError:
            return False

    def fire_lifecycle_event(self, new_state):
        """Called when instance's state changes.

        Args:
            new_state (str): The new state of the instance.
        """
        _logger.info("HazelcastClient %s is %s", __version__, new_state)
        for on_state_change in six.itervalues(self._listeners):
            if on_state_change:
                try:
                    on_state_change(new_state)
                except:
                    _logger.exception("Exception in lifecycle listener")



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.codec.custom.raft_group_id_codec import RaftGroupIdCodec
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x090200
_REQUEST_MESSAGE_TYPE = 590336
# hex: 0x090201
_RESPONSE_MESSAGE_TYPE = 590337

_REQUEST_RETURN_VALUE_TYPE_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_RETURN_VALUE_TYPE_OFFSET + INT_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE


def encode_request(group_id, name, function, return_value_type):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_RETURN_VALUE_TYPE_OFFSET, return_value_type)
    RaftGroupIdCodec.encode(buf, group_id)
    StringCodec.encode(buf, name)
    DataCodec.encode(buf, function, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)



from hazelcast.protocol.codec import (
    transactional_set_add_codec,
    transactional_set_remove_codec,
    transactional_set_size_codec,
)
from hazelcast.proxy.base import TransactionalProxy
from hazelcast.util import check_not_none, thread_id


class TransactionalSet(TransactionalProxy):
    """Transactional implementation of :class:`~hazelcast.proxy.set.Set`."""

    def add(self, item):
        """Transactional implementation of :func:`Set.add(item) <hazelcast.proxy.set.Set.add>`

        Args:
            item: The new item to be added.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if item is added successfully, ``False`` otherwise.
        """
        check_not_none(item, "item can't be none")
        item_data = self._to_data(item)
        request = transactional_set_add_codec.encode_request(
            self.name, self.transaction.id, thread_id(), item_data
        )
        return self._invoke(request, transactional_set_add_codec.decode_response)

    def remove(self, item):
        """Transactional implementation of :func:`Set.remove(item) <hazelcast.proxy.set.Set.remove>`

        Args:
            item: The specified item to be deleted.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if item is remove successfully, ``False`` otherwise.
        """
        check_not_none(item, "item can't be none")
        item_data = self._to_data(item)
        request = transactional_set_remove_codec.encode_request(
            self.name, self.transaction.id, thread_id(), item_data
        )
        return self._invoke(request, transactional_set_remove_codec.decode_response)

    def size(self):
        """Transactional implementation of :func:`Set.size() <hazelcast.proxy.set.Set.size>`

        Returns:
            hazelcast.future.Future[int]: Size of the set.
        """
        request = transactional_set_size_codec.encode_request(
            self.name, self.transaction.id, thread_id()
        )
        return self._invoke(request, transactional_set_size_codec.decode_response)



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import EntryListCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x0D1100
_REQUEST_MESSAGE_TYPE = 856320
# hex: 0x0D1101
_RESPONSE_MESSAGE_TYPE = 856321

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(name):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    msg.next_frame()
    return EntryListCodec.decode(msg, DataCodec.decode, DataCodec.decode)



from hazelcast.six import string_types
from hazelcast.core import Address


class TokenProvider(object):
    """TokenProvider is a base class for token providers."""

    def token(self, address=None):
        # type: (TokenProvider, Address) -> bytes
        """Returns a token to be used for token-based authentication.

        Args:
            address (hazelcast.core.Address): Connected address for the member.

        Returns:
            bytes: token as a bytes object.
        """
        pass


class BasicTokenProvider(TokenProvider):
    """BasicTokenProvider sends the given token to the authentication endpoint."""

    def __init__(self, token=""):
        if isinstance(token, string_types):
            self._token = token.encode("utf-8")
        elif isinstance(token, bytes):
            self._token = token
        else:
            raise TypeError("token must be either a str or bytes object")

    def token(self, address=None):
        return self._token



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.builtin import CodecUtil

# hex: 0x012E00
_REQUEST_MESSAGE_TYPE = 77312
# hex: 0x012E01
_RESPONSE_MESSAGE_TYPE = 77313

_REQUEST_THREAD_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_THREAD_ID_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(name, entry_processor, key, thread_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_THREAD_ID_OFFSET, thread_id)
    StringCodec.encode(buf, name)
    DataCodec.encode(buf, entry_processor)
    DataCodec.encode(buf, key, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    msg.next_frame()
    return CodecUtil.decode_nullable(msg, DataCodec.decode)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x051500
_REQUEST_MESSAGE_TYPE = 333056
# hex: 0x051501
_RESPONSE_MESSAGE_TYPE = 333057

_REQUEST_FROM_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_TO_OFFSET = _REQUEST_FROM_OFFSET + INT_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_TO_OFFSET + INT_SIZE_IN_BYTES


def encode_request(name, _from, to):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_FROM_OFFSET, _from)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_TO_OFFSET, to)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    msg.next_frame()
    return ListMultiFrameCodec.decode(msg, DataCodec.decode)



from hazelcast.protocol.codec import (
    list_add_all_codec,
    list_add_all_with_index_codec,
    list_add_codec,
    list_add_listener_codec,
    list_add_with_index_codec,
    list_clear_codec,
    list_compare_and_remove_all_codec,
    list_compare_and_retain_all_codec,
    list_contains_all_codec,
    list_contains_codec,
    list_get_all_codec,
    list_get_codec,
    list_index_of_codec,
    list_is_empty_codec,
    list_iterator_codec,
    list_last_index_of_codec,
    list_list_iterator_codec,
    list_remove_codec,
    list_remove_listener_codec,
    list_remove_with_index_codec,
    list_set_codec,
    list_size_codec,
    list_sub_codec,
)
from hazelcast.proxy.base import PartitionSpecificProxy, ItemEvent, ItemEventType
from hazelcast.util import check_not_none, ImmutableLazyDataList


class List(PartitionSpecificProxy):
    """Concurrent, distributed implementation of List.

    The Hazelcast List is not a partitioned data-structure. So all the content of the List is stored in a single
    machine (and in the backup). So the List will not scale by adding more members in the cluster.
    """

    def add(self, item):
        """Adds the specified item to the end of this list.

        Args:
            item: the specified item to be appended to this list.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if item is added, ``False`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)
        request = list_add_codec.encode_request(self.name, element_data)
        return self._invoke(request, list_add_codec.decode_response)

    def add_at(self, index, item):
        """Adds the specified item at the specific position in this list.
        Element in this position and following elements are shifted to the right, if any.

        Args:
            index (int): The specified index to insert the item.
            item: The specified item to be inserted.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)

        request = list_add_with_index_codec.encode_request(self.name, index, element_data)
        return self._invoke(request)

    def add_all(self, items):
        """Adds all of the items in the specified collection to the end of this list.

        The order of new elements is determined by the specified collection's iterator.

        Args:
            items (list): The specified collection which includes the elements to be added to list.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this call changed the list, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = list_add_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, list_add_all_codec.decode_response)

    def add_all_at(self, index, items):
        """Adds all of the elements in the specified collection into this list at the specified position.

        Elements in this positions and following elements are shifted to the right, if any.
        The order of new elements is determined by the specified collection's iterator.

        Args:
            index (int): The specified index at which the first element of specified collection is added.
            items (list): The specified collection which includes the elements to be added to list.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this call changed the list, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = list_add_all_with_index_codec.encode_request(self.name, index, data_items)
        return self._invoke(request, list_add_all_with_index_codec.decode_response)

    def add_listener(self, include_value=False, item_added_func=None, item_removed_func=None):
        """Adds an item listener for this list. Listener will be notified for all list add/remove events.

        Args:
            include_value (bool): Whether received events include the updated item or not.
            item_added_func (function):  To be called when an item is added to this list.
            item_removed_func (function): To be called when an item is deleted from this list.

        Returns:
            hazelcast.future.Future[str]: A registration id which is used as a key to remove the listener.
        """
        request = list_add_listener_codec.encode_request(self.name, include_value, self._is_smart)

        def handle_event_item(item, uuid, event_type):
            item = item if include_value else None
            member = self._context.cluster_service.get_member(uuid)

            item_event = ItemEvent(self.name, item, event_type, member, self._to_object)
            if event_type == ItemEventType.ADDED:
                if item_added_func:
                    item_added_func(item_event)
            else:
                if item_removed_func:
                    item_removed_func(item_event)

        return self._register_listener(
            request,
            lambda r: list_add_listener_codec.decode_response(r),
            lambda reg_id: list_remove_listener_codec.encode_request(self.name, reg_id),
            lambda m: list_add_listener_codec.handle(m, handle_event_item),
        )

    def clear(self):
        """Clears the list.

        List will be empty with this call.

        Returns:
            hazelcast.future.Future[None]:
        """
        request = list_clear_codec.encode_request(self.name)
        return self._invoke(request)

    def contains(self, item):
        """Determines whether this list contains the specified item or not.

        Args:
            item: The specified item.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the specified item exists in this list, ``False`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)

        request = list_contains_codec.encode_request(self.name, item_data)
        return self._invoke(request, list_contains_codec.decode_response)

    def contains_all(self, items):
        """Determines whether this list contains all of the items in specified collection or not.

        Args:
          items (list): The specified collection which includes the items to be searched.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if all of the items in specified collection
            exist in this list, ``False`` otherwise.
        """
        check_not_none(items, "Items can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "item can't be None")
            data_items.append(self._to_data(item))

        request = list_contains_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, list_contains_all_codec.decode_response)

    def get(self, index):
        """Returns the item which is in the specified position in this list.

        Args:
          index (int): the specified index of the item to be returned.

        Returns:
            hazelcast.future.Future[any]: the item in the specified position in this list.
        """

        def handler(message):
            return self._to_object(list_get_codec.decode_response(message))

        request = list_get_codec.encode_request(self.name, index)
        return self._invoke(request, handler)

    def get_all(self):
        """Returns all of the items in this list.

        Returns:
            hazelcast.future.Future[list]: All of the items in this list.
        """

        def handler(message):
            return ImmutableLazyDataList(
                list_get_all_codec.decode_response(message), self._to_object
            )

        request = list_get_all_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def iterator(self):
        """Returns an iterator over the elements in this list in proper sequence, same with ``get_all``.

        Returns:
            hazelcast.future.Future[list]: All of the items in this list.
        """

        def handler(message):
            return ImmutableLazyDataList(
                list_iterator_codec.decode_response(message), self._to_object
            )

        request = list_iterator_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def index_of(self, item):
        """Returns the first index of specified items's occurrences in this list.

        If specified item is not present in this list, returns -1.

        Args:
            item: The specified item to be searched for.

        Returns:
             hazelcast.future.Future[int]: The first index of specified items's occurrences,
             ``-1`` if item is not present in this list.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)

        request = list_index_of_codec.encode_request(self.name, item_data)
        return self._invoke(request, list_index_of_codec.decode_response)

    def is_empty(self):
        """Determines whether this list is empty or not.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the list contains no elements, ``False`` otherwise.
        """

        request = list_is_empty_codec.encode_request(self.name)
        return self._invoke(request, list_is_empty_codec.decode_response)

    def last_index_of(self, item):
        """Returns the last index of specified items's occurrences in this list.

        If specified item is not present in this list, returns -1.

        Args:
            item: The specified item to be searched for.

        Returns:
            hazelcast.future.Future[int]: The last index of specified items's occurrences,
            ``-1`` if item is not present in this list.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)

        request = list_last_index_of_codec.encode_request(self.name, item_data)
        return self._invoke(request, list_last_index_of_codec.decode_response)

    def list_iterator(self, index=0):
        """Returns a list iterator of the elements in this list.

        If an index is provided, iterator starts from this index.

        Args:
            index: (int), index of first element to be returned from the list iterator.

        Returns:
            hazelcast.future.Future[list]: List of the elements in this list.
        """

        def handler(message):
            return ImmutableLazyDataList(
                list_list_iterator_codec.decode_response(message), self._to_object
            )

        request = list_list_iterator_codec.encode_request(self.name, index)
        return self._invoke(request, handler)

    def remove(self, item):
        """Removes the specified element's first occurrence from the list if it exists in this list.

        Args:
            item: The specified element.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the specified element is present in this list,
            ``False`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)

        request = list_remove_codec.encode_request(self.name, item_data)
        return self._invoke(request, list_remove_codec.decode_response)

    def remove_at(self, index):
        """Removes the item at the specified position in this list.

        Element in this position and following elements are shifted to the left, if any.

        Args:
            index (int): Index of the item to be removed.

        Returns:
            hazelcast.future.Future[any]: The item previously at the specified index.
        """

        def handler(message):
            return self._to_object(list_remove_with_index_codec.decode_response(message))

        request = list_remove_with_index_codec.encode_request(self.name, index)
        return self._invoke(request, handler)

    def remove_all(self, items):
        """Removes all of the elements that is present in the specified collection from this list.

        Args:
            items (list): The specified collection.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this list changed as a result of the call, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = list_compare_and_remove_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, list_compare_and_remove_all_codec.decode_response)

    def remove_listener(self, registration_id):
        """Removes the specified item listener.

        Returns silently if the specified listener was not added before.

        Args:
            registration_id (str): Id of the listener to be deleted.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the item listener is removed, ``False`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def retain_all(self, items):
        """Retains only the items that are contained in the specified collection.

        It means, items which are not present in the specified collection are removed from this list.

        Args:
            items (list): Collections which includes the elements to be retained in this list.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this list changed as a result of the call, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = list_compare_and_retain_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, list_compare_and_retain_all_codec.decode_response)

    def size(self):
        """Returns the number of elements in this list.

        Returns:
            hazelcast.future.Future[int]: Number of elements in this list.
        """
        request = list_size_codec.encode_request(self.name)
        return self._invoke(request, list_size_codec.decode_response)

    def set_at(self, index, item):
        """Replaces the specified element with the element at the specified position in this list.

        Args:
            index (int): Index of the item to be replaced.
            item: Item to be stored.

        Returns:
            hazelcast.future.Future[any]: the previous item in the specified index.
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)

        def handler(message):
            return self._to_object(list_set_codec.decode_response(message))

        request = list_set_codec.encode_request(self.name, index, element_data)
        return self._invoke(request, handler)

    def sub_list(self, from_index, to_index):
        """Returns a sublist from this list, from from_index(inclusive) to to_index(exclusive).

        The returned list is backed by this list, so non-structural changes in the returned list are reflected in this
        list, and vice-versa.

        Args:
            from_index (int): The start point(inclusive) of the sub_list.
            to_index (int): The end point(exclusive) of the sub_list.

        Returns:
            hazelcast.future.Future[list]: A view of the specified range within this list.
        """

        def handler(message):
            return ImmutableLazyDataList(list_sub_codec.decode_response(message), self._to_object)

        request = list_sub_codec.encode_request(self.name, from_index, to_index)
        return self._invoke(request, handler)



import uuid

from hazelcast.errors import OperationTimeoutError
from hazelcast.protocol.codec import (
    count_down_latch_await_codec,
    count_down_latch_get_round_codec,
    count_down_latch_count_down_codec,
    count_down_latch_get_count_codec,
    count_down_latch_try_set_count_codec,
)
from hazelcast.proxy.cp import BaseCPProxy
from hazelcast.util import to_millis, check_true, check_is_number, check_is_int


class CountDownLatch(BaseCPProxy):
    """A distributed, concurrent countdown latch data structure.

    CountDownLatch is a cluster-wide synchronization aid
    that allows one or more callers to wait until a set of operations being
    performed in other callers completes.

    CountDownLatch count can be reset using ``try_set_count()`` method after
    a countdown has finished but not during an active count. This allows
    the same latch instance to be reused.

    There is no ``await_latch()`` method to do an unbound wait since this is undesirable
    in a distributed application: for example, a cluster can split or the master
    and replicas could all die. In most cases, it is best to configure
    an explicit timeout so you have the ability to deal with these situations.

    All of the API methods in the CountDownLatch offer the exactly-once
    execution semantics. For instance, even if a ``count_down()`` call is
    internally retried because of crashed Hazelcast member, the counter
    value is decremented only once.
    """

    def await_latch(self, timeout):
        """Causes the current thread to wait until the latch has counted down to
        zero, or an exception is thrown, or the specified waiting time elapses.

        If the current count is zero then this method returns ``True``.

        If the current count is greater than zero, then the current
        thread becomes disabled for thread scheduling purposes and lies
        dormant until one of the following things happen:

        - The count reaches zero due to invocations of the ``count_down()`` method
        - This CountDownLatch instance is destroyed
        - The countdown owner becomes disconnected
        - The specified waiting time elapses

        If the count reaches zero, then the method returns with the
        value ``True``.

        If the specified waiting time elapses then the value ``False``
        is returned.  If the time is less than or equal to zero, the method
        will not wait at all.

        Args:
            timeout (int): The maximum time to wait in seconds

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the count reached zero,
            ``False`` if the waiting time elapsed before the count reached zero
        Raises:
            IllegalStateError: If the Hazelcast instance was shut down while waiting.
        """
        check_is_number(timeout)
        timeout = max(0, timeout)
        invocation_uuid = uuid.uuid4()
        codec = count_down_latch_await_codec
        request = codec.encode_request(
            self._group_id, self._object_name, invocation_uuid, to_millis(timeout)
        )
        return self._invoke(request, codec.decode_response)

    def count_down(self):
        """Decrements the count of the latch, releasing all waiting threads if
        the count reaches zero.

        If the current count is greater than zero, then it is decremented.
        If the new count is zero:

        - All waiting threads are re-enabled for thread scheduling purposes
        - Countdown owner is set to ``None``.

        If the current count equals zero, then nothing happens.

        Returns:
            hazelcast.future.Future[None]:
        """
        invocation_uuid = uuid.uuid4()

        def handler(f):
            return self._do_count_down(f.result(), invocation_uuid)

        return self._get_round().continue_with(handler)

    def get_count(self):
        """Returns the current count.

        Returns:
            hazelcast.future.Future[int]: The current count.
        """
        codec = count_down_latch_get_count_codec
        request = codec.encode_request(self._group_id, self._object_name)
        return self._invoke(request, codec.decode_response)

    def try_set_count(self, count):
        """Sets the count to the given value if the current count is zero.

        If count is not zero, then this method does nothing and returns
        ``False``.

        Args:
            count (int): The number of times ``count_down()`` must be invoked
                before callers can pass through ``await_latch()``.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the new count was set,
            ``False`` if the current count is not zero.
        """
        check_is_int(count)
        check_true(count > 0, "Count must be positive")
        codec = count_down_latch_try_set_count_codec
        request = codec.encode_request(self._group_id, self._object_name, count)
        return self._invoke(request, codec.decode_response)

    def _do_count_down(self, expected_round, invocation_uuid):
        def handler(f):
            try:
                f.result()
            except OperationTimeoutError:
                # we can retry safely because the retry is idempotent
                return self._do_count_down(expected_round, invocation_uuid)

        return self._request_count_down(expected_round, invocation_uuid).continue_with(handler)

    def _get_round(self):
        codec = count_down_latch_get_round_codec
        request = codec.encode_request(self._group_id, self._object_name)
        return self._invoke(request, codec.decode_response)

    def _request_count_down(self, expected_round, invocation_uuid):
        codec = count_down_latch_count_down_codec
        request = codec.encode_request(
            self._group_id, self._object_name, invocation_uuid, expected_round
        )
        return self._invoke(request)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.builtin import CodecUtil
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.builtin import LongArrayCodec

# hex: 0x170900
_REQUEST_MESSAGE_TYPE = 1509632
# hex: 0x170901
_RESPONSE_MESSAGE_TYPE = 1509633

_REQUEST_START_SEQUENCE_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_MIN_COUNT_OFFSET = _REQUEST_START_SEQUENCE_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_MAX_COUNT_OFFSET = _REQUEST_MIN_COUNT_OFFSET + INT_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_MAX_COUNT_OFFSET + INT_SIZE_IN_BYTES
_RESPONSE_READ_COUNT_OFFSET = RESPONSE_HEADER_SIZE
_RESPONSE_NEXT_SEQ_OFFSET = _RESPONSE_READ_COUNT_OFFSET + INT_SIZE_IN_BYTES


def encode_request(name, start_sequence, min_count, max_count, filter):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_START_SEQUENCE_OFFSET, start_sequence)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_MIN_COUNT_OFFSET, min_count)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_MAX_COUNT_OFFSET, max_count)
    StringCodec.encode(buf, name)
    CodecUtil.encode_nullable(buf, filter, DataCodec.encode, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    initial_frame = msg.next_frame()
    response = dict()
    response["read_count"] = FixSizedTypesCodec.decode_int(initial_frame.buf, _RESPONSE_READ_COUNT_OFFSET)
    response["next_seq"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_NEXT_SEQ_OFFSET)
    response["items"] = ListMultiFrameCodec.decode(msg, DataCodec.decode)
    response["item_seqs"] = CodecUtil.decode_nullable(msg, LongArrayCodec.decode)
    return response



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.protocol import RaftGroupId
from hazelcast.protocol.builtin import StringCodec

_SEED_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_SEED_DECODE_OFFSET = 0
_ID_ENCODE_OFFSET = _SEED_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_ID_DECODE_OFFSET = _SEED_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_INITIAL_FRAME_SIZE = _ID_ENCODE_OFFSET + LONG_SIZE_IN_BYTES - 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS


class RaftGroupIdCodec(object):
    @staticmethod
    def encode(buf, raft_group_id, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _SEED_ENCODE_OFFSET, raft_group_id.seed)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _ID_ENCODE_OFFSET, raft_group_id.id)
        buf.extend(initial_frame_buf)
        StringCodec.encode(buf, raft_group_id.name)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        seed = FixSizedTypesCodec.decode_long(initial_frame.buf, _SEED_DECODE_OFFSET)
        id = FixSizedTypesCodec.decode_long(initial_frame.buf, _ID_DECODE_OFFSET)
        name = StringCodec.decode(msg)
        CodecUtil.fast_forward_to_end_frame(msg)
        return RaftGroupId(name, seed, id)



import uuid

from hazelcast import six
from hazelcast.serialization.base import BaseSerializationService
from hazelcast.serialization.objects import (
    CanonicalizingHashSet,
    IdentifiedAddress,
    IdentifiedMapEntry,
    ReliableTopicMessage,
)
from hazelcast.serialization.portable.classdef import FieldType
from hazelcast.serialization.portable.context import PortableContext
from hazelcast.serialization.portable.serializer import PortableSerializer
from hazelcast.serialization.serializer import *

DEFAULT_OUT_BUFFER_SIZE = 4 * 1024


def default_partition_strategy(key):
    if hasattr(key, "get_partition_key"):
        return key.get_partition_key()
    return None


class SerializationServiceV1(BaseSerializationService):
    def __init__(
        self,
        config,
        version=1,
        global_partition_strategy=default_partition_strategy,
        output_buffer_size=DEFAULT_OUT_BUFFER_SIZE,
    ):
        super(SerializationServiceV1, self).__init__(
            version,
            global_partition_strategy,
            output_buffer_size,
            config.is_big_endian,
            config.default_int_type,
        )
        self._portable_context = PortableContext(self, config.portable_version)
        self.register_class_definitions(
            config.class_definitions, config.check_class_definition_errors
        )
        self._registry._portable_serializer = PortableSerializer(
            self._portable_context, config.portable_factories
        )

        # merge configured factories with built in ones
        factories = self._get_builtin_identified_factories()
        factories.update(config.data_serializable_factories)
        self._registry._data_serializer = IdentifiedDataSerializer(factories)
        self._register_constant_serializers()

        # Register Custom Serializers
        for _type, custom_serializer in six.iteritems(config.custom_serializers):
            self._registry.safe_register_serializer(custom_serializer(), _type)

        # Register Global Serializer
        global_serializer = config.global_serializer
        if global_serializer:
            self._registry._global_serializer = global_serializer()

    def _get_builtin_identified_factories(self):
        return {
            ReliableTopicMessage.FACTORY_ID: {
                ReliableTopicMessage.CLASS_ID: ReliableTopicMessage,
            },
            CanonicalizingHashSet.FACTORY_ID: {
                CanonicalizingHashSet.CLASS_ID: CanonicalizingHashSet,
            },
            IdentifiedMapEntry.FACTORY_ID: {
                IdentifiedMapEntry.CLASS_ID: IdentifiedMapEntry,
            },
            IdentifiedAddress.FACTORY_ID: {
                IdentifiedAddress.CLASS_ID: IdentifiedAddress,
            },
        }

    def _register_constant_serializers(self):
        self._registry.register_constant_serializer(self._registry._null_serializer, type(None))
        self._registry.register_constant_serializer(self._registry._data_serializer)
        self._registry.register_constant_serializer(self._registry._portable_serializer)
        self._registry.register_constant_serializer(ByteSerializer())
        self._registry.register_constant_serializer(BooleanSerializer(), bool)
        self._registry.register_constant_serializer(CharSerializer())
        self._registry.register_constant_serializer(ShortSerializer())
        self._registry.register_constant_serializer(IntegerSerializer())
        self._registry.register_constant_serializer(LongSerializer())
        self._registry.register_constant_serializer(FloatSerializer())
        self._registry.register_constant_serializer(DoubleSerializer(), float)
        self._registry.register_constant_serializer(UuidSerializer(), uuid.UUID)
        self._registry.register_constant_serializer(StringSerializer())
        # Arrays of primitives and String
        self._registry.register_constant_serializer(ByteArraySerializer(), bytearray)
        self._registry.register_constant_serializer(BooleanArraySerializer())
        self._registry.register_constant_serializer(CharArraySerializer())
        self._registry.register_constant_serializer(ShortArraySerializer())
        self._registry.register_constant_serializer(IntegerArraySerializer())
        self._registry.register_constant_serializer(LongArraySerializer())
        self._registry.register_constant_serializer(FloatArraySerializer())
        self._registry.register_constant_serializer(DoubleArraySerializer())
        self._registry.register_constant_serializer(StringArraySerializer())
        # EXTENSIONS
        self._registry.register_constant_serializer(DateTimeSerializer(), datetime)
        self._registry.register_constant_serializer(BigIntegerSerializer())
        self._registry.register_constant_serializer(JavaClassSerializer())
        self._registry.register_constant_serializer(ArraySerializer())
        self._registry.register_constant_serializer(ArrayListSerializer(), list)
        self._registry.register_constant_serializer(LinkedListSerializer())
        self._registry.register_constant_serializer(
            HazelcastJsonValueSerializer(), HazelcastJsonValue
        )

        self._registry.safe_register_serializer(self._registry._python_serializer)

    def register_class_definitions(self, class_definitions, check_error):
        factories = dict()
        for cd in class_definitions:
            factory_id = cd.factory_id
            class_defs = factories.get(factory_id, None)
            if class_defs is None:
                class_defs = dict()
                factories[factory_id] = class_defs

            class_id = cd.class_id
            if class_id in class_defs:
                raise HazelcastSerializationError(
                    "Duplicate registration found for class-id: %s" % class_id
                )
            class_defs[class_id] = cd

        for cd in class_definitions:
            self.register_class_definition(cd, factories, check_error)

    def register_class_definition(self, class_definition, factories, check_error):
        field_names = class_definition.get_field_names()
        for field_name in field_names:
            fd = class_definition.get_field(field_name)
            if fd.field_type == FieldType.PORTABLE or fd.field_type == FieldType.PORTABLE_ARRAY:
                factory_id = fd.factory_id
                class_id = fd.class_id
                class_defs = factories.get(factory_id, None)
                if class_defs is not None:
                    nested_cd = class_defs.get(class_id, None)
                    if nested_cd is not None:
                        self.register_class_definition(nested_cd, factories, check_error)
                        self._portable_context.register_class_definition(nested_cd)
                        continue

                if check_error:
                    raise HazelcastSerializationError(
                        "Could not find registered ClassDefinition for factory-id: %s, class-id: %s"
                        % (factory_id, class_id)
                    )

        self._portable_context.register_class_definition(class_definition)



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.codec.custom.paging_predicate_holder_codec import PagingPredicateHolderCodec
from hazelcast.protocol.builtin import EntryListCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.codec.custom.anchor_data_list_holder_codec import AnchorDataListHolderCodec

# hex: 0x013600
_REQUEST_MESSAGE_TYPE = 79360
# hex: 0x013601
_RESPONSE_MESSAGE_TYPE = 79361

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(name, predicate):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    StringCodec.encode(buf, name)
    PagingPredicateHolderCodec.encode(buf, predicate, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    msg.next_frame()
    response = dict()
    response["response"] = EntryListCodec.decode(msg, DataCodec.decode, DataCodec.decode)
    response["anchor_data_list"] = AnchorDataListHolderCodec.decode(msg)
    return response



import time
import uuid

from hazelcast.errors import SessionExpiredError, WaitKeyCancelledError, IllegalStateError
from hazelcast.future import ImmediateFuture, ImmediateExceptionFuture
from hazelcast.protocol.codec import (
    semaphore_init_codec,
    semaphore_acquire_codec,
    semaphore_available_permits_codec,
    semaphore_drain_codec,
    semaphore_change_codec,
    semaphore_release_codec,
)
from hazelcast.proxy.cp import SessionAwareCPProxy, BaseCPProxy
from hazelcast.util import check_not_negative, check_true, thread_id, to_millis

_DRAIN_SESSION_ACQ_COUNT = 1024
"""
Since a proxy does not know how many permits will be drained on
the Raft group, it uses this constant to increment its local session
acquire count. Then, it adjusts the local session acquire count after
the drain response is returned.
"""

_NO_SESSION_ID = -1


class Semaphore(BaseCPProxy):
    """A linearizable, distributed semaphore.

    Semaphores are often used to restrict the number of callers that can access
    some physical or logical resource.

    Semaphore is a cluster-wide counting semaphore. Conceptually, it maintains
    a set of permits. Each ``acquire()`` blocks if necessary until a permit
    is available, and then takes it. Dually, each ``release()`` adds a
    permit, potentially releasing a blocking acquirer. However, no actual permit
    objects are used; the semaphore just keeps a count of the number available
    and acts accordingly.

    Hazelcast's distributed semaphore implementation guarantees that callers
    invoking any of the ``acquire()`` methods are selected to
    obtain permits in the order of their invocations (first-in-first-out; FIFO).
    Note that FIFO ordering implies the order which the primary replica of an
    Semaphore receives these acquire requests. Therefore, it is
    possible for one member to invoke ``acquire()`` before another member,
    but its request hits the primary replica after the other member.

    This class also provides convenient ways to work with multiple permits at once.
    Beware of the increased risk of indefinite postponement when using the
    multiple-permit acquire. If permits are released one by one, a caller waiting
    for one permit will acquire it before a caller waiting for multiple permits
    regardless of the call order.

    Correct usage of a semaphore is established by programming convention
    in the application.

    It works on top of the Raft consensus algorithm. It offers linearizability during crash
    failures and network partitions. It is CP with respect to the CAP principle.
    If a network partition occurs, it remains available on at most one side of
    the partition.

    It has 2 variations:

    - The default implementation accessed via ``cp_subsystem`` is session-aware. In this
      one, when a caller makes its very first ``acquire()`` call, it starts
      a new CP session with the underlying CP group. Then, liveliness of the
      caller is tracked via this CP session. When the caller fails, permits
      acquired by this caller are automatically and safely released. However,
      the session-aware version comes with a limitation, that is, a client cannot
      release permits before acquiring them first. In other words, a client can release
      only the permits it has acquired earlier. It means, you can acquire a permit
      from one thread and release it from another thread using the same Hazelcast client,
      but not different instances of Hazelcast client. You can use the session-aware
      CP Semaphore implementation by disabling JDK compatibility via ``jdk-compatible``
      server-side setting. Although the session-aware implementation has a minor
      difference to the JDK Semaphore, we think it is a better fit for distributed
      environments because of its safe auto-cleanup mechanism for acquired permits.
    - The second implementation offered by ``cp_subsystem`` is sessionless. This
      implementation does not perform auto-cleanup of acquired permits on failures.
      Acquired permits are not bound to threads and permits can be released without
      acquiring first. However, you need to handle failed permit owners on your own.
      If a Hazelcast server or a client fails while holding some permits, they will not be
      automatically released. You can use the sessionless CP Semaphore implementation
      by enabling JDK compatibility via ``jdk-compatible`` server-side setting.

    There is a subtle difference between the lock and semaphore abstractions.
    A lock can be assigned to at most one endpoint at a time, so we have a total
    order among its holders. However, permits of a semaphore can be assigned to
    multiple endpoints at a time, which implies that we may not have a total
    order among permit holders. In fact, permit holders are partially ordered.
    For this reason, the fencing token approach, which is explained in
    :class:`~hazelcast.proxy.cp.fenced_lock.FencedLock`, does not work for the
    semaphore abstraction. Moreover, each permit is an independent entity.
    Multiple permit acquires and reentrant lock acquires of a single endpoint are
    not equivalent. The only case where a semaphore behaves like a lock is the
    binary case, where the semaphore has only 1 permit. In this case, the semaphore
    works like a non-reentrant lock.

    All of the API methods in the new CP Semaphore implementation offer
    the exactly-once execution semantics for the session-aware version.
    For instance, even if a ``release()`` call is internally retried
    because of a crashed Hazelcast member, the permit is released only once.
    However, this guarantee is not given for the sessionless, a.k.a,
    JDK-compatible CP Semaphore.
    """

    def init(self, permits):
        """Tries to initialize this Semaphore instance with the given permit count.

        Args:
            permits (int): The given permit count.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the initialization succeeds,
            ``False`` if already initialized.

        Raises:
            AssertionError: If the ``permits`` is negative.
        """
        check_not_negative(permits, "Permits must be non-negative")
        codec = semaphore_init_codec
        request = codec.encode_request(self._group_id, self._object_name, permits)
        return self._invoke(request, codec.decode_response)

    def acquire(self, permits=1):
        """Acquires the given number of permits if they are available,
        and returns immediately, reducing the number of available permits
        by the given amount.

        If insufficient permits are available then the result of the returned
        future is not set until one of the following things happens:

        - Some other caller invokes one of the ``release``
          methods for this semaphore, the current caller is next to be assigned
          permits and the number of available permits satisfies this request,
        - This Semaphore instance is destroyed

        Args:
            permits (int): Optional number of permits to acquire; defaults to ``1``
                when not specified

        Returns:
            hazelcast.future.Future[None]:

        Raises:
            AssertionError: If the ``permits`` is not positive.
        """
        raise NotImplementedError("acquire")

    def available_permits(self):
        """Returns the current number of permits currently available in this semaphore.

        This method is typically used for debugging and testing purposes.

        Returns:
            hazelcast.future.Future[int]: The number of permits available in this semaphore.
        """
        codec = semaphore_available_permits_codec
        request = codec.encode_request(self._group_id, self._object_name)
        return self._invoke(request, codec.decode_response)

    def drain_permits(self):
        """Acquires and returns all permits that are available at invocation time.

        Returns:
            hazelcast.future.Future[int]: The number of permits drained.
        """
        raise NotImplementedError("drain_permits")

    def reduce_permits(self, reduction):
        """Reduces the number of available permits by the indicated amount.

        This method differs from ``acquire`` as it does not block until permits
        become available. Similarly, if the caller has acquired some permits,
        they are not released with this call.

        Args:
            reduction (int): The number of permits to reduce.

        Returns:
            hazelcast.future.Future[None]:

        Raises:
             AssertionError: If the ``reduction`` is negative.
        """
        check_not_negative(reduction, "Reduction must be non-negative")
        if reduction == 0:
            return ImmediateFuture(None)

        return self._do_change_permits(-reduction)

    def increase_permits(self, increase):
        """Increases the number of available permits by the indicated amount.

        If there are some callers waiting for permits to become available, they
        will be notified. Moreover, if the caller has acquired some permits,
        they are not released with this call.

        Args:
            increase (int): The number of permits to increase.

        Returns:
            hazelcast.future.Future[None]:

        Raises:
            AssertionError: If ``increase`` is negative.
        """
        check_not_negative(increase, "Increase must be non-negative")
        if increase == 0:
            return ImmediateFuture(None)

        return self._do_change_permits(increase)

    def release(self, permits=1):
        """Releases the given number of permits and increases the number of
        available permits by that amount.

        If some callers in the cluster are blocked for acquiring permits,
        they will be notified.

        If the underlying Semaphore implementation is non-JDK-compatible
        (configured via ``jdk-compatible`` server-side setting), then a
        client can only release a permit which it has acquired before.
        In other words, a client cannot release a permit without acquiring
        it first.

        Otherwise, which means the underlying implementation is JDK compatible
        (configured via ``jdk-compatible`` server-side setting), there is no
        requirement that a client that releases a permit must have acquired
        that permit by calling one of the ``acquire()`` methods. A client can
        freely release a permit without acquiring it first. In this case,
        correct usage of a semaphore is established by programming convention
        in the application.

        Args:
            permits (int): Optional number of permits to release; defaults to ``1``
                when not specified.

        Returns:
            hazelcast.future.Future[None]:

        Raises:
            AssertionError: If the ``permits`` is not positive.
            IllegalStateError: if the Semaphore is non-JDK-compatible and the caller
                does not have a permit
        """
        raise NotImplementedError("release")

    def try_acquire(self, permits=1, timeout=0):
        """Acquires the given number of permits and returns ``True``, if they
        become available during the given waiting time.

        If permits are acquired, the number of available permits in the Semaphore
        instance is also reduced by the given amount.

        If no sufficient permits are available, then the result of the returned
        future is not set until one of the following things happens:

        - Permits are released by other callers, the current caller is next to
          be assigned permits and the number of available permits satisfies this
          request
        - The specified waiting time elapses

        Args:
            permits (int): The number of permits to acquire; defaults to ``1``
                when not specified.
            timeout (int): Optional timeout in seconds to wait for the permits;
                when it's not specified the operation will return
                immediately after the acquire attempt

        Returns:
            hazelcast.future.Future[bool]: ``True`` if all permits were acquired,
            ``False`` if the waiting time elapsed before all permits could be
            acquired

        Raises:
            AssertionError: If the ``permits`` is not positive.
        """
        raise NotImplementedError("try_acquire")

    def _do_change_permits(self, permits):
        raise NotImplementedError("_do_change_permits")


class SessionAwareSemaphore(Semaphore, SessionAwareCPProxy):
    def acquire(self, permits=1):
        check_true(permits > 0, "Permits must be positive")
        current_thread_id = thread_id()
        invocation_uuid = uuid.uuid4()
        return self._do_acquire(current_thread_id, invocation_uuid, permits)

    def drain_permits(self):
        current_thread_id = thread_id()
        invocation_uuid = uuid.uuid4()
        return self._do_drain(current_thread_id, invocation_uuid)

    def release(self, permits=1):
        check_true(permits > 0, "Permits must be positive")
        session_id = self._get_session_id()
        if session_id == _NO_SESSION_ID:
            return ImmediateExceptionFuture(self._new_illegal_state_error())

        current_thread_id = thread_id()
        invocation_uuid = uuid.uuid4()

        def check_response(response):
            try:
                response.result()
            except SessionExpiredError as e:
                self._invalidate_session(session_id)
                raise self._new_illegal_state_error(e)
            finally:
                self._release_session(session_id, permits)

        return self._request_release(
            session_id, current_thread_id, invocation_uuid, permits
        ).continue_with(check_response)

    def try_acquire(self, permits=1, timeout=0):
        check_true(permits > 0, "Permits must be positive")
        timeout = max(0, timeout)
        current_thread_id = thread_id()
        invocation_uuid = uuid.uuid4()
        return self._do_try_acquire(current_thread_id, invocation_uuid, permits, timeout)

    def _do_acquire(self, current_thread_id, invocation_uuid, permits):
        def do_acquire_once(session_id):
            session_id = session_id.result()

            def check_response(response):
                try:
                    response.result()
                except SessionExpiredError:
                    self._invalidate_session(session_id)
                    return self._do_acquire(current_thread_id, invocation_uuid, permits)
                except WaitKeyCancelledError:
                    self._release_session(session_id, permits)
                    error = IllegalStateError(
                        'Semaphore("%s") not acquired because the acquire call on the CP '
                        "group is cancelled, possibly because of another indeterminate call "
                        "from the same thread." % self._object_name
                    )
                    raise error
                except Exception as e:
                    self._release_session(session_id, permits)
                    raise e

            return self._request_acquire(
                session_id, current_thread_id, invocation_uuid, permits, -1
            ).continue_with(check_response)

        return self._acquire_session(permits).continue_with(do_acquire_once)

    def _do_drain(self, current_thread_id, invocation_uuid):
        def do_drain_once(session_id):
            session_id = session_id.result()

            def check_count(count):
                try:
                    count = count.result()
                    self._release_session(session_id, _DRAIN_SESSION_ACQ_COUNT - count)
                    return count
                except SessionExpiredError:
                    self._invalidate_session(session_id)
                    return self._do_drain(current_thread_id, invocation_uuid)
                except Exception as e:
                    self._release_session(session_id, _DRAIN_SESSION_ACQ_COUNT)
                    raise e

            return self._request_drain(
                session_id, current_thread_id, invocation_uuid
            ).continue_with(check_count)

        return self._acquire_session(_DRAIN_SESSION_ACQ_COUNT).continue_with(do_drain_once)

    def _do_change_permits(self, delta):
        current_thread_id = thread_id()
        invocation_uuid = uuid.uuid4()

        def do_change_permits_once(session_id):
            session_id = session_id.result()

            def check_response(response):
                try:
                    response.result()
                except SessionExpiredError as e:
                    self._invalidate_session(session_id)
                    raise self._new_illegal_state_error(e)
                finally:
                    self._release_session(session_id)

            return self._request_change(
                session_id, current_thread_id, invocation_uuid, delta
            ).continue_with(check_response)

        return self._acquire_session().continue_with(do_change_permits_once)

    def _do_try_acquire(self, current_thread_id, invocation_uuid, permits, timeout):
        start = time.time()

        def do_try_acquire_once(session_id):
            session_id = session_id.result()

            def check_response(response):
                try:
                    acquired = response.result()
                    if not acquired:
                        self._release_session(session_id, permits)
                    return acquired
                except SessionExpiredError:
                    self._invalidate_session(session_id)
                    remaining_timeout = timeout - (time.time() - start)
                    if remaining_timeout <= 0:
                        return False
                    return self._do_try_acquire(
                        current_thread_id, invocation_uuid, permits, remaining_timeout
                    )
                except WaitKeyCancelledError:
                    self._release_session(session_id, permits)
                    return False
                except Exception as e:
                    self._release_session(session_id, permits)
                    raise e

            return self._request_acquire(
                session_id, current_thread_id, invocation_uuid, permits, timeout
            ).continue_with(check_response)

        return self._acquire_session(permits).continue_with(do_try_acquire_once)

    def _new_illegal_state_error(self, cause=None):
        error = IllegalStateError(
            'Semaphore["%s"] has no valid session!' % self._object_name, cause
        )
        return error

    def _request_acquire(self, session_id, current_thread_id, invocation_uuid, permits, timeout):
        codec = semaphore_acquire_codec
        if timeout > 0:
            timeout = to_millis(timeout)

        request = codec.encode_request(
            self._group_id,
            self._object_name,
            session_id,
            current_thread_id,
            invocation_uuid,
            permits,
            timeout,
        )
        return self._invoke(request, codec.decode_response)

    def _request_drain(self, session_id, current_thread_id, invocation_uuid):
        codec = semaphore_drain_codec
        request = codec.encode_request(
            self._group_id, self._object_name, session_id, current_thread_id, invocation_uuid
        )
        return self._invoke(request, codec.decode_response)

    def _request_change(self, session_id, current_thread_id, invocation_uuid, delta):
        codec = semaphore_change_codec
        request = codec.encode_request(
            self._group_id, self._object_name, session_id, current_thread_id, invocation_uuid, delta
        )
        return self._invoke(request)

    def _request_release(self, session_id, current_thread_id, invocation_uuid, permits):
        codec = semaphore_release_codec
        request = codec.encode_request(
            self._group_id,
            self._object_name,
            session_id,
            current_thread_id,
            invocation_uuid,
            permits,
        )
        return self._invoke(request)


class SessionlessSemaphore(Semaphore):
    def __init__(self, context, group_id, service_name, proxy_name, object_name):
        super(SessionlessSemaphore, self).__init__(
            context, group_id, service_name, proxy_name, object_name
        )
        self._session_manager = context.proxy_session_manager

    def acquire(self, permits=1):
        check_true(permits > 0, "Permits must be positive")

        def handler(f):
            f.result()
            return None

        return (
            self._get_thread_id()
            .continue_with(self._do_try_acquire, permits, -1)
            .continue_with(handler)
        )

    def drain_permits(self):
        return self._get_thread_id().continue_with(self._do_drain_permits)

    def release(self, permits=1):
        check_true(permits > 0, "Permits must be positive")
        invocation_uuid = uuid.uuid4()
        return self._get_thread_id().continue_with(self._request_release, invocation_uuid, permits)

    def try_acquire(self, permits=1, timeout=0):
        check_true(permits > 0, "Permits must be positive")
        timeout = max(0, timeout)
        return self._get_thread_id().continue_with(self._do_try_acquire, permits, timeout)

    def _do_try_acquire(self, global_thread_id, permits, timeout):
        global_thread_id = global_thread_id.result()
        invocation_uuid = uuid.uuid4()
        return self._request_acquire(
            global_thread_id, invocation_uuid, permits, timeout
        ).continue_with(self._check_acquire_response)

    def _do_drain_permits(self, global_thread_id):
        global_thread_id = global_thread_id.result()
        invocation_uuid = uuid.uuid4()
        codec = semaphore_drain_codec
        request = codec.encode_request(
            self._group_id, self._object_name, _NO_SESSION_ID, global_thread_id, invocation_uuid
        )
        return self._invoke(request, codec.decode_response)

    def _do_change_permits(self, permits):
        invocation_uuid = uuid.uuid4()
        return self._get_thread_id().continue_with(self._request_change, invocation_uuid, permits)

    def _request_acquire(self, global_thread_id, invocation_uuid, permits, timeout):
        codec = semaphore_acquire_codec
        if timeout > 0:
            timeout = to_millis(timeout)

        request = codec.encode_request(
            self._group_id,
            self._object_name,
            _NO_SESSION_ID,
            global_thread_id,
            invocation_uuid,
            permits,
            timeout,
        )
        return self._invoke(request, codec.decode_response)

    def _request_change(self, global_thread_id, invocation_uuid, permits):
        global_thread_id = global_thread_id.result()
        codec = semaphore_change_codec
        request = codec.encode_request(
            self._group_id,
            self._object_name,
            _NO_SESSION_ID,
            global_thread_id,
            invocation_uuid,
            permits,
        )
        return self._invoke(request)

    def _request_release(self, global_thread_id, invocation_uuid, permits):
        global_thread_id = global_thread_id.result()
        codec = semaphore_release_codec
        request = codec.encode_request(
            self._group_id,
            self._object_name,
            _NO_SESSION_ID,
            global_thread_id,
            invocation_uuid,
            permits,
        )
        return self._invoke(request)

    def _check_acquire_response(self, response):
        try:
            return response.result()
        except WaitKeyCancelledError:
            error = IllegalStateError(
                'Semaphore("%s") not acquired because the acquire call on the '
                "CP group is cancelled, possibly because of another indeterminate "
                "call from the same thread." % self._object_name
            )
            raise error

    def _get_thread_id(self):
        return self._session_manager.get_or_create_unique_thread_id(self._group_id)
